from collections.abc import Mapping

import pytest
from dirty_equals import IsStrictDict
from inline_snapshot import snapshot

from pydantic_core import SchemaValidator, ValidationError, core_schema


def test_list():
    v = SchemaValidator(
        core_schema.list_schema(
            core_schema.tuple_positional_schema([core_schema.int_schema(), core_schema.int_schema()]),
        )
    )
    assert v.validate_python([[1, 2], [3, 4]]) == [(1, 2), (3, 4)]
    assert v.validate_python([[1, 2], [3, 4]], allow_partial=True) == [(1, 2), (3, 4)]
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python([[1, 2], 'wrong'])
    assert exc_info.value.errors(include_url=False) == snapshot(
        [
            {
                'type': 'tuple_type',
                'loc': (1,),
                'msg': 'Input should be a valid tuple',
                'input': 'wrong',
            }
        ]
    )
    assert v.validate_python([[1, 2], 'wrong'], allow_partial=True) == [(1, 2)]
    assert v.validate_python([[1, 2], []], allow_partial=True) == [(1, 2)]
    assert v.validate_python([[1, 2], [3]], allow_partial=True) == [(1, 2)]
    assert v.validate_python([[1, 2], [3, 'x']], allow_partial=True) == [(1, 2)]
    with pytest.raises(ValidationError, match='Input should be a valid tuple'):
        v.validate_python([[1, 2], 'wrong', [3, 4]])
    with pytest.raises(ValidationError, match='Input should be a valid tuple'):
        v.validate_python([[1, 2], 'wrong', 'wrong'])
    assert v.validate_json(b'[[1, 2], [3, 4]]', allow_partial=True) == [(1, 2), (3, 4)]
    assert v.validate_json(b'[[1, 2], [3,', allow_partial=True) == [(1, 2)]


@pytest.mark.parametrize('collection_type', [core_schema.set_schema, core_schema.frozenset_schema])
def test_set_frozenset(collection_type):
    v = SchemaValidator(
        collection_type(
            core_schema.tuple_positional_schema([core_schema.int_schema(), core_schema.int_schema()]),
        )
    )
    assert v.validate_python([[1, 2], [3, 4]]) == snapshot({(1, 2), (3, 4)})
    assert v.validate_python([[1, 2], [3, 4]], allow_partial=True) == snapshot({(1, 2), (3, 4)})
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python([[1, 2], 'wrong'])
    assert exc_info.value.errors(include_url=False) == snapshot(
        [
            {
                'type': 'tuple_type',
                'loc': (1,),
                'msg': 'Input should be a valid tuple',
                'input': 'wrong',
            }
        ]
    )
    assert v.validate_python([[1, 2], 'wrong'], allow_partial=True) == snapshot({(1, 2)})
    assert v.validate_python([[1, 2], [3, 4], 'wrong'], allow_partial=True) == snapshot({(1, 2), (3, 4)})
    assert v.validate_python([[1, 2], []], allow_partial=True) == snapshot({(1, 2)})
    assert v.validate_python([[1, 2], [3]], allow_partial=True) == snapshot({(1, 2)})
    assert v.validate_python([[1, 2], [3, 'x']], allow_partial=True) == snapshot({(1, 2)})
    with pytest.raises(ValidationError, match='Input should be a valid tuple'):
        v.validate_python([[1, 2], 'wrong', [3, 4]])
    with pytest.raises(ValidationError, match='Input should be a valid tuple'):
        v.validate_python([[1, 2], 'wrong', 'wrong'])


class MyMapping(Mapping):
    def __init__(self, d):
        self._d = d

    def __getitem__(self, key):
        return self._d[key]

    def __iter__(self):
        return iter(self._d)

    def __len__(self):
        return len(self._d)


def test_dict():
    v = SchemaValidator(core_schema.dict_schema(core_schema.int_schema(), core_schema.int_schema()))
    assert v.validate_python({'1': 2, 3: '4'}) == snapshot({1: 2, 3: 4})
    assert v.validate_python({'1': 2, 3: '4'}, allow_partial=True) == snapshot({1: 2, 3: 4})
    assert v.validate_python(MyMapping({'1': 2, 3: '4'}), allow_partial=True) == snapshot({1: 2, 3: 4})
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python({'1': 2, 3: 'wrong'})
    assert exc_info.value.errors(include_url=False) == snapshot(
        [
            {
                'type': 'int_parsing',
                'loc': (3,),
                'msg': 'Input should be a valid integer, unable to parse string as an integer',
                'input': 'wrong',
            }
        ]
    )
    assert v.validate_python({'1': 2, 3: 'x'}, allow_partial=True) == snapshot({1: 2})
    assert v.validate_python(MyMapping({'1': 2, 3: 'x'}), allow_partial=True) == snapshot({1: 2})
    assert v.validate_python({'1': 2, 3: 4, 5: '6', 7: 'x'}, allow_partial=True) == snapshot({1: 2, 3: 4, 5: 6})
    with pytest.raises(ValidationError, match='Input should be a valid integer'):
        v.validate_python({'1': 2, 3: 4, 5: 'x', 7: '8'})
    with pytest.raises(ValidationError, match='Input should be a valid integer'):
        v.validate_python({'1': 2, 3: 4, 5: 'x', 7: 'x'})
    with pytest.raises(ValidationError, match='Input should be a valid integer'):
        v.validate_python({'1': 2, 3: 4, 'x': 6})


def test_dict_list():
    v = SchemaValidator(
        core_schema.dict_schema(core_schema.int_schema(), core_schema.list_schema(core_schema.int_schema(ge=10)))
    )
    assert v.validate_python({'1': [20, 30], 3: [40, '50']}, allow_partial=True) == snapshot({1: [20, 30], 3: [40, 50]})
    assert v.validate_python({'1': [20, 30], 3: [40, 5]}, allow_partial=True) == snapshot({1: [20, 30], 3: [40]})

    with pytest.raises(ValidationError, match=r'1\.1\s+Input should be greater than or equal to 10'):
        v.validate_python({'1': [20, 3], 3: [40, 50]}, allow_partial=True)


def test_partial_typed_dict():
    v = SchemaValidator(
        core_schema.typed_dict_schema(
            {
                'a': core_schema.typed_dict_field(core_schema.int_schema(gt=10)),
                'b': core_schema.typed_dict_field(core_schema.int_schema(gt=10)),
                'c': core_schema.typed_dict_field(core_schema.int_schema(gt=10)),
            },
            total=False,
        )
    )

    assert v.validate_python({'a': 11, 'b': '12', 'c': 13}) == snapshot(IsStrictDict(a=11, b=12, c=13))
    assert v.validate_python({'a': 11, 'c': 13, 'b': '12'}) == snapshot(IsStrictDict(a=11, b=12, c=13))
    assert v.validate_python(MyMapping({'a': 11, 'c': 13, 'b': '12'})) == snapshot(IsStrictDict(a=11, b=12, c=13))

    assert v.validate_python({'a': 11, 'b': '12', 'c': 13}, allow_partial=True) == snapshot({'a': 11, 'b': 12, 'c': 13})
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python({'a': 11, 'b': '12', 'c': 1})
    assert exc_info.value.errors(include_url=False) == snapshot(
        [
            {
                'type': 'greater_than',
                'loc': ('c',),
                'msg': 'Input should be greater than 10',
                'input': 1,
                'ctx': {'gt': 10},
            }
        ]
    )
    assert v.validate_python({'a': 11, 'b': '12', 'c': 1}, allow_partial=True) == snapshot(IsStrictDict(a=11, b=12))
    assert v.validate_python(MyMapping({'a': 11, 'b': '12', 'c': 1}), allow_partial=True) == snapshot(
        IsStrictDict(a=11, b=12)
    )
    assert v.validate_python({'a': 11, 'c': 13, 'b': 1}, allow_partial=True) == snapshot(IsStrictDict(a=11, c=13))
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python({'a': 11, 'c': 1, 'b': 12}, allow_partial=True)
    assert exc_info.value.errors(include_url=False) == snapshot(
        [
            {
                'type': 'greater_than',
                'loc': ('c',),
                'msg': 'Input should be greater than 10',
                'input': 1,
                'ctx': {'gt': 10},
            }
        ]
    )
    with pytest.raises(ValidationError, match=r'c\s+Input should be greater than 10'):
        v.validate_python(MyMapping({'a': 11, 'c': 1, 'b': 12}), allow_partial=True)

    # validate strings
    assert v.validate_strings({'a': '11', 'b': '22'}) == snapshot({'a': 11, 'b': 22})
    with pytest.raises(ValidationError, match='Input should be greater than 10'):
        v.validate_strings({'a': '11', 'b': '2'})
    assert v.validate_strings({'a': '11', 'b': '2'}, allow_partial=True) == snapshot({'a': 11})

    assert v.validate_json(b'{"b": "12", "a": 11, "c": 13}', allow_partial=True) == IsStrictDict(a=11, b=12, c=13)
    assert v.validate_json(b'{"b": "12", "a": 11, "c": 13', allow_partial=True) == IsStrictDict(a=11, b=12, c=13)
    assert v.validate_json(b'{"a": 11, "b": "12", "c": 1', allow_partial=True) == IsStrictDict(a=11, b=12)
    assert v.validate_json(b'{"a": 11, "b": "12", "c":', allow_partial=True) == IsStrictDict(a=11, b=12)
    assert v.validate_json(b'{"a": 11, "b": "12", "c"', allow_partial=True) == IsStrictDict(a=11, b=12)
    assert v.validate_json(b'{"a": 11, "b": "12", "c', allow_partial=True) == IsStrictDict(a=11, b=12)
    assert v.validate_json(b'{"a": 11, "b": "12", "', allow_partial=True) == IsStrictDict(a=11, b=12)
    assert v.validate_json(b'{"a": 11, "b": "12", ', allow_partial=True) == IsStrictDict(a=11, b=12)
    assert v.validate_json(b'{"a": 11, "b": "12",', allow_partial=True) == IsStrictDict(a=11, b=12)
    assert v.validate_json(b'{"a": 11, "b": "12"', allow_partial=True) == IsStrictDict(a=11, b=12)


def test_non_partial_typed_dict():
    v = SchemaValidator(
        core_schema.typed_dict_schema(
            {
                'a': core_schema.typed_dict_field(core_schema.int_schema(gt=10)),
                'b': core_schema.typed_dict_field(core_schema.int_schema(gt=10), required=True),
                'c': core_schema.typed_dict_field(core_schema.int_schema(gt=10)),
            },
            total=False,
        )
    )

    assert v.validate_python({'a': 11, 'b': '12', 'c': 13}) == snapshot({'a': 11, 'b': 12, 'c': 13})
    with pytest.raises(ValidationError, match='Input should be greater than 10'):
        v.validate_python({'a': 11, 'b': '12', 'c': 1})
    assert v.validate_python({'a': 11, 'b': '12', 'c': 1}, allow_partial=True) == snapshot({'a': 11, 'b': 12})
    with pytest.raises(ValidationError, match=r'b\s+Field required'):
        v.validate_python({'a': 11, 'c': 12}, allow_partial=True)
    with pytest.raises(ValidationError, match=r'b\s+Input should be greater than 10'):
        v.validate_python({'a': 11, 'c': 12, 'b': 1}, allow_partial=True)


def test_double_nested():
    v = SchemaValidator(
        core_schema.typed_dict_schema(
            {
                'a': core_schema.typed_dict_field(core_schema.int_schema(gt=10)),
                'b': core_schema.typed_dict_field(
                    core_schema.list_schema(
                        core_schema.dict_schema(core_schema.str_schema(), core_schema.int_schema(ge=10))
                    )
                ),
            },
            total=False,
        )
    )
    assert v.validate_python({'a': 11, 'b': [{'a': 10, 'b': 20}, {'a': 30, 'b': 40}]}) == snapshot(
        {'a': 11, 'b': [{'a': 10, 'b': 20}, {'a': 30, 'b': 40}]}
    )
    assert v.validate_python({'a': 11, 'b': [{'a': 10, 'b': 20}, {'a': 30, 'b': 4}]}, allow_partial=True) == snapshot(
        {'a': 11, 'b': [{'a': 10, 'b': 20}, {'a': 30}]}
    )
    assert v.validate_python({'a': 11, 'b': [{'a': 10, 'b': 20}, {'a': 30, 123: 4}]}, allow_partial=True) == snapshot(
        {'a': 11, 'b': [{'a': 10, 'b': 20}]}
    )
    # the first element of the list is invalid, so the whole list is invalid
    assert v.validate_python({'a': 11, 'b': [{'a': 10, 'b': 2}, {'a': 30}]}, allow_partial=True) == snapshot({'a': 11})
    with pytest.raises(ValidationError, match=r'b\.0\.b\s+Input should be greater than or equal to 10'):
        v.validate_python({'b': [{'a': 10, 'b': 2}, {'a': 30}], 'a': 11}, allow_partial=True)

    with pytest.raises(ValidationError, match=r'b\.1\.a\s+Input should be greater than or equal to 10'):
        v.validate_python({'b': [{'a': 10, 'b': 20}, {'a': 3}], 'a': 11}, allow_partial=True)

    assert v.validate_python({'a': 11, 'b': [{'a': 1, 'b': 20}, {'a': 3, 'b': 40}]}, allow_partial=True) == snapshot(
        {'a': 11}
    )
    json = b'{"a": 11, "b": [{"a": 10, "b": 20}, {"a": 30, "b": 40}]}'
    assert v.validate_json(json, allow_partial=True) == snapshot(
        {'a': 11, 'b': [{'a': 10, 'b': 20}, {'a': 30, 'b': 40}]}
    )
    for i in range(1, len(json)):
        value = v.validate_json(json[:i], allow_partial=True)
        assert isinstance(value, dict)


def test_tuple_list():
    """Tuples don't support partial, so behaviour should be disabled."""
    v = SchemaValidator(
        core_schema.tuple_positional_schema(
            [core_schema.list_schema(core_schema.int_schema()), core_schema.int_schema()]
        )
    )
    assert v.validate_python([['1', '2'], '3'], allow_partial=True) == snapshot(([1, 2], 3))
    with pytest.raises(ValidationError, match=r'1\s+Input should be a valid integer'):
        v.validate_python([['1', '2'], 'x'], allow_partial=True)
    with pytest.raises(ValidationError, match=r'0\.1\s+Input should be a valid integer'):
        v.validate_python([['1', 'x'], '2'], allow_partial=True)


def test_dataclass():
    """Tuples don't support partial, so behaviour should be disabled."""

    schema = core_schema.dataclass_args_schema(
        'MyDataclass',
        [
            core_schema.dataclass_field(name='a', schema=core_schema.str_schema(), kw_only=False),
            core_schema.dataclass_field(
                name='b', schema=core_schema.list_schema(core_schema.str_schema(min_length=2)), kw_only=False
            ),
        ],
    )
    v = SchemaValidator(schema)
    assert v.validate_python({'a': 'x', 'b': ['ab', 'cd']}) == snapshot(({'a': 'x', 'b': ['ab', 'cd']}, None))
    assert v.validate_python({'a': 'x', 'b': ['ab', 'cd']}, allow_partial=True) == snapshot(
        ({'a': 'x', 'b': ['ab', 'cd']}, None)
    )
    with pytest.raises(ValidationError, match=r'b\.1\s+String should have at least 2 characters'):
        v.validate_python({'a': 'x', 'b': ['ab', 'c']}, allow_partial=True)


def test_nullable():
    v = SchemaValidator(core_schema.nullable_schema(core_schema.list_schema(core_schema.str_schema(min_length=2))))

    assert v.validate_python(None, allow_partial=True) is None
    assert v.validate_python(['ab', 'cd'], allow_partial=True) == ['ab', 'cd']
    assert v.validate_python(['ab', 'c'], allow_partial=True) == ['ab']
    assert v.validate_json('["ab", "cd"]', allow_partial=True) == ['ab', 'cd']
    assert v.validate_json('["ab", "cd', allow_partial=True) == ['ab']
    assert v.validate_json('["ab", "cd', allow_partial='trailing-strings') == ['ab', 'cd']
    assert v.validate_json('["ab", "c', allow_partial=True) == ['ab']
    assert v.validate_json('["ab", "c', allow_partial='trailing-strings') == ['ab']


@pytest.mark.parametrize(
    'json_nested_type', [None, core_schema.dict_schema(core_schema.str_schema(), core_schema.int_schema())]
)
def test_json(json_nested_type):
    v = SchemaValidator(core_schema.list_schema(core_schema.json_schema(json_nested_type)))

    assert v.validate_python(['{"a": 1}', '{"b": 2}']) == snapshot([{'a': 1}, {'b': 2}])
    assert v.validate_python(['{"a": 1}', '{"b": 2}'], allow_partial=True) == snapshot([{'a': 1}, {'b': 2}])
    assert v.validate_python(['{"a": 1}', 'xxx'], allow_partial=True) == snapshot([{'a': 1}])
    assert v.validate_python(['{"a": 1}', '{"b": 2'], allow_partial=True) == snapshot([{'a': 1}, {'b': 2}])
    assert v.validate_json('["{\\"a\\": 1}", "{\\"b\\": 2}', allow_partial='trailing-strings') == snapshot(
        [{'a': 1}, {'b': 2}]
    )
    assert v.validate_json('["{\\"a\\": 1}", "{\\"b\\": 2', allow_partial='trailing-strings') == snapshot(
        [{'a': 1}, {'b': 2}]
    )


def test_json_trailing_strings():
    v = SchemaValidator(core_schema.list_schema(core_schema.json_schema()))
    assert v.validate_python(['{"a": 1}', '{"b": "x'], allow_partial=True) == snapshot([{'a': 1}, {}])
    assert v.validate_python(['{"a": 1}', '{"b": "x'], allow_partial='trailing-strings') == snapshot(
        [{'a': 1}, {'b': 'x'}]
    )

    assert v.validate_json('["{\\"a\\": 1}", "{\\"b\\": 2}"]') == snapshot([{'a': 1}, {'b': 2}])
    assert v.validate_json('["{\\"a\\": 1}", "{\\"b\\": 2, \\"c\\": \\"x', allow_partial=True) == snapshot([{'a': 1}])
    assert v.validate_json(
        '["{\\"a\\": 1}", "{\\"b\\": 2, \\"c\\": \\"x', allow_partial='trailing-strings'
    ) == snapshot([{'a': 1}, {'b': 2, 'c': 'x'}])



import re

import pytest
from dirty_equals import HasRepr, IsStr

from pydantic_core import CoreConfig, SchemaValidator, ValidationError
from pydantic_core import core_schema as cs

from ..conftest import Err, PyAndJson


@pytest.mark.parametrize(
    'input_value,expected',
    [
        ([1, 2, 3], [1, 2, 3]),
        ([1, 2, '3'], [1, 2, 3]),
        ({1: 2, 3: 4}, [1, 3]),
        ('123', [1, 2, 3]),
        (5, Err('[type=iterable_type, input_value=5, input_type=int]')),
        ([1, 'wrong'], Err("[type=int_parsing, input_value='wrong', input_type=str]")),
    ],
    ids=repr,
)
def test_generator_json_int(py_and_json: PyAndJson, input_value, expected):
    v = py_and_json({'type': 'generator', 'items_schema': {'type': 'int'}})
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            list(v.validate_test(input_value))

    else:
        assert list(v.validate_test(input_value)) == expected


@pytest.mark.parametrize(
    'config,input_str',
    (
        (CoreConfig(), 'type=iterable_type, input_value=5, input_type=int'),
        (CoreConfig(hide_input_in_errors=False), 'type=iterable_type, input_value=5, input_type=int'),
        (CoreConfig(hide_input_in_errors=True), 'type=iterable_type'),
    ),
)
def test_generator_json_hide_input(py_and_json: PyAndJson, config, input_str):
    v = py_and_json({'type': 'generator', 'items_schema': {'type': 'int'}}, config)
    with pytest.raises(ValidationError, match=re.escape(f'[{input_str}]')):
        list(v.validate_test(5))


@pytest.mark.parametrize(
    'input_value,expected',
    [
        ([1, 2, 3], [1, 2, 3]),
        ([1, 2, '3'], [1, 2, '3']),
        ({'1': 2, '3': 4}, ['1', '3']),
        ('123', ['1', '2', '3']),
        (5, Err('[type=iterable_type, input_value=5, input_type=int]')),
        ([1, 'wrong'], [1, 'wrong']),
    ],
    ids=repr,
)
def test_generator_json_any(py_and_json: PyAndJson, input_value, expected):
    v = py_and_json({'type': 'generator'})
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            list(v.validate_test(input_value))

    else:
        assert list(v.validate_test(input_value)) == expected


def test_error_index(py_and_json: PyAndJson):
    v = py_and_json({'type': 'generator', 'items_schema': {'type': 'int'}})
    gen = v.validate_test(['wrong'])
    assert gen.index == 0
    with pytest.raises(ValidationError) as exc_info:
        next(gen)
    assert gen.index == 1
    # insert_assert(exc_info.value.errors(include_url=False))
    assert exc_info.value.title == 'ValidatorIterator'
    assert str(exc_info.value).startswith('1 validation error for ValidatorIterator\n')
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'int_parsing',
            'loc': (0,),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'wrong',
        }
    ]
    gen = v.validate_test([1, 2, 3, 'wrong', 4])
    assert gen.index == 0
    assert next(gen) == 1
    assert gen.index == 1
    assert next(gen) == 2
    assert gen.index == 2
    assert next(gen) == 3
    assert gen.index == 3
    with pytest.raises(ValidationError) as exc_info:
        next(gen)
    assert gen.index == 4
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'int_parsing',
            'loc': (3,),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'wrong',
        }
    ]
    assert next(gen) == 4
    assert gen.index == 5


def test_too_long(py_and_json: PyAndJson):
    v = py_and_json({'type': 'generator', 'items_schema': {'type': 'int'}, 'max_length': 2})
    assert list(v.validate_test([1])) == [1]
    assert list(v.validate_test([1, 2])) == [1, 2]
    with pytest.raises(ValidationError) as exc_info:
        list(v.validate_test([1, 2, 3]))
    # insert_assert(exc_info.value.errors(include_url=False))
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'too_long',
            'loc': (),
            'msg': 'Generator should have at most 2 items after validation, not more',
            'input': [1, 2, 3],
            'ctx': {'field_type': 'Generator', 'max_length': 2, 'actual_length': None},
        }
    ]


def test_too_short(py_and_json: PyAndJson):
    v = py_and_json({'type': 'generator', 'items_schema': {'type': 'int'}, 'min_length': 2})
    assert list(v.validate_test([1, 2, 3])) == [1, 2, 3]
    assert list(v.validate_test([1, 2])) == [1, 2]
    with pytest.raises(ValidationError) as exc_info:
        list(v.validate_test([1]))
    # insert_assert(exc_info.value.errors(include_url=False))
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'too_short',
            'loc': (),
            'msg': 'Generator should have at least 2 items after validation, not 1',
            'input': [1],
            'ctx': {'field_type': 'Generator', 'min_length': 2, 'actual_length': 1},
        }
    ]


def gen():
    yield 1
    yield 2
    yield 3


def test_generator_too_long():
    v = SchemaValidator(cs.generator_schema(items_schema=cs.int_schema(), max_length=2))

    validating_iterator = v.validate_python(gen())

    # Ensure the error happens at exactly the right step:
    assert next(validating_iterator) == 1
    assert next(validating_iterator) == 2
    with pytest.raises(ValidationError) as exc_info:
        next(validating_iterator)

    errors = exc_info.value.errors(include_url=False)
    # insert_assert(errors)
    assert errors == [
        {
            'type': 'too_long',
            'loc': (),
            'input': HasRepr(IsStr(regex='<generator object gen at .+>')),
            'msg': 'Generator should have at most 2 items after validation, not more',
            'ctx': {'field_type': 'Generator', 'max_length': 2, 'actual_length': None},
        }
    ]


def test_generator_too_short():
    v = SchemaValidator(cs.generator_schema(items_schema=cs.int_schema(), min_length=4))

    validating_iterator = v.validate_python(gen())

    # Ensure the error happens at exactly the right step:
    assert next(validating_iterator) == 1
    assert next(validating_iterator) == 2
    assert next(validating_iterator) == 3
    with pytest.raises(ValidationError) as exc_info:
        next(validating_iterator)

    errors = exc_info.value.errors(include_url=False)
    # insert_assert(errors)
    assert errors == [
        {
            'type': 'too_short',
            'input': HasRepr(IsStr(regex='<generator object gen at .+>')),
            'loc': (),
            'msg': 'Generator should have at least 4 items after validation, not 3',
            'ctx': {'field_type': 'Generator', 'min_length': 4, 'actual_length': 3},
        }
    ]



import base64
import json
from enum import Enum

import pytest

from pydantic_core import PydanticSerializationError, SchemaSerializer, core_schema, to_json


def test_bytes():
    s = SchemaSerializer(core_schema.bytes_schema())
    assert s.to_python(b'foobar') == b'foobar'
    assert s.to_python('emoji 💩'.encode()) == 'emoji 💩'.encode()
    assert s.to_json(b'foobar') == b'"foobar"'
    assert s.to_python(b'foobar', mode='json') == 'foobar'

    json_emoji = s.to_json('emoji 💩'.encode())
    # note! serde_json serializes unicode characters differently
    assert json_emoji == b'"emoji \xf0\x9f\x92\xa9"'
    assert json.loads(json_emoji) == 'emoji 💩'


def test_bytes_invalid_all():
    s = SchemaSerializer(core_schema.bytes_schema())
    assert s.to_python(b'\x81') == b'\x81'

    msg = 'Error serializing to JSON: invalid utf-8 sequence of 1 bytes from index 0'
    with pytest.raises(PydanticSerializationError, match=msg):
        s.to_json(b'\x81')


def test_bytes_invalid_cpython():
    # PyO3/pyo3#2770 is now fixed
    s = SchemaSerializer(core_schema.bytes_schema())

    with pytest.raises(UnicodeDecodeError, match="'utf-8' codec can't decode byte 0x81 in position 0: invalid utf-8"):
        s.to_python(b'\x81', mode='json')


def test_bytes_dict_key():
    s = SchemaSerializer(core_schema.dict_schema(core_schema.bytes_schema(), core_schema.int_schema()))
    assert s.to_python({b'foobar': 123}) == {b'foobar': 123}
    assert s.to_python({b'foobar': 123}, mode='json') == {'foobar': 123}
    assert s.to_json({b'foobar': 123}) == b'{"foobar":123}'


def test_bytes_fallback():
    s = SchemaSerializer(core_schema.bytes_schema())
    with pytest.warns(
        UserWarning,
        match=r'Expected `bytes` - serialized value may not be as expected \[input_value=123, input_type=int\]',
    ):
        assert s.to_python(123) == 123
    with pytest.warns(
        UserWarning,
        match=r'Expected `bytes` - serialized value may not be as expected \[input_value=123, input_type=int\]',
    ):
        assert s.to_python(123, mode='json') == 123
    with pytest.warns(
        UserWarning,
        match=r'Expected `bytes` - serialized value may not be as expected \[input_value=123, input_type=int\]',
    ):
        assert s.to_json(123) == b'123'
    with pytest.warns(
        UserWarning,
        match=r"Expected `bytes` - serialized value may not be as expected \[input_value='foo', input_type=str\]",
    ):
        assert s.to_json('foo') == b'"foo"'


class BytesSubclass(bytes):
    pass


class BasicClass:
    pass


class BytesMixin(bytes, BasicClass):
    pass


class BytesEnum(bytes, Enum):
    foo = b'foo-value'
    bar = b'bar-value'


@pytest.mark.parametrize('schema_type', ['bytes', 'any'])
@pytest.mark.parametrize(
    'input_value,expected_json',
    [(BytesSubclass(b'foo'), 'foo'), (BytesMixin(b'foo'), 'foo'), (BytesEnum.foo, 'foo-value')],
)
def test_subclass_bytes(schema_type, input_value, expected_json):
    s = SchemaSerializer({'type': schema_type})
    v = s.to_python(input_value)
    assert v == input_value
    assert type(v) == type(input_value)

    v = s.to_python(input_value, mode='json')
    assert v == expected_json
    assert type(v) == str

    assert s.to_json(input_value) == json.dumps(expected_json).encode('utf-8')


def test_bytes_base64():
    s = SchemaSerializer(core_schema.bytes_schema(), {'ser_json_bytes': 'base64'})
    assert s.to_python(b'foobar') == b'foobar'

    assert s.to_json(b'foobar') == b'"Zm9vYmFy"'
    assert s.to_python(b'foobar', mode='json') == 'Zm9vYmFy'
    assert base64.b64decode(s.to_python(b'foobar', mode='json').encode()) == b'foobar'

    # with padding
    assert s.to_json(b'foo bar') == b'"Zm9vIGJhcg=="'
    assert s.to_python(b'foo bar', mode='json') == 'Zm9vIGJhcg=='
    assert base64.b64decode(s.to_python(b'foo bar', mode='json').encode()) == b'foo bar'


def test_bytes_hex():
    s = SchemaSerializer(core_schema.bytes_schema(), {'ser_json_bytes': 'hex'})
    assert s.to_python(b'\xff\xff') == b'\xff\xff'
    assert s.to_json(b'\xff\xff') == b'"ffff"'
    assert s.to_python(b'\xff\xff', mode='json') == 'ffff' == b'\xff\xff'.hex()


def test_bytes_base64_dict_key():
    s = SchemaSerializer(core_schema.dict_schema(core_schema.bytes_schema()), {'ser_json_bytes': 'base64'})

    assert s.to_python({b'foo bar': 123}, mode='json') == {'Zm9vIGJhcg==': 123}
    assert s.to_json({b'foo bar': 123}) == b'{"Zm9vIGJhcg==":123}'


def test_any_bytes_base64():
    s = SchemaSerializer(core_schema.any_schema(), {'ser_json_bytes': 'base64'})
    assert s.to_python(b'foobar') == b'foobar'

    assert s.to_json(b'foobar') == b'"Zm9vYmFy"'
    assert s.to_json({b'foobar': 123}) == b'{"Zm9vYmFy":123}'
    assert s.to_python({b'foobar': 123}, mode='json') == {'Zm9vYmFy': 123}


class BasicModel:
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


def test_bytes_mode_set_via_model_config_not_serializer_config():
    s = SchemaSerializer(
        core_schema.model_schema(
            BasicModel,
            core_schema.model_fields_schema(
                {
                    'foo': core_schema.model_field(core_schema.bytes_schema()),
                }
            ),
            config=core_schema.CoreConfig(ser_json_bytes='base64'),
        )
    )

    bm = BasicModel(foo=b'foobar')
    assert s.to_python(bm) == {'foo': b'foobar'}
    assert s.to_json(bm) == b'{"foo":"Zm9vYmFy"}'
    assert s.to_python(bm, mode='json') == {'foo': 'Zm9vYmFy'}

    # assert doesn't override serializer config
    # in V3, we can change the serialization settings provided to to_json to override model config settings,
    # but that'd be a breaking change
    BasicModel.__pydantic_serializer__ = s
    assert to_json(bm, bytes_mode='utf8') == b'{"foo":"Zm9vYmFy"}'

    assert to_json({'foo': b'some bytes'}, bytes_mode='base64') == b'{"foo":"c29tZSBieXRlcw=="}'
    assert to_json({'bar': bm}, bytes_mode='base64') == b'{"bar":{"foo":"Zm9vYmFy"}}'



import pytest

from pydantic_core import SchemaError, SchemaValidator
from pydantic_core import core_schema as cs


def test_build_non_default_follows_default() -> None:
    with pytest.raises(SchemaError, match="Required parameter 'b' follows parameter with default"):
        SchemaValidator(
            schema=cs.arguments_v3_schema(
                [
                    cs.arguments_v3_parameter(
                        name='a',
                        schema=cs.with_default_schema(schema=cs.int_schema(), default_factory=lambda: 42),
                        mode='positional_or_keyword',
                    ),
                    cs.arguments_v3_parameter(name='b', schema=cs.int_schema(), mode='positional_or_keyword'),
                ]
            )
        )


def test_duplicate_parameter_name() -> None:
    with pytest.raises(SchemaError, match="Duplicate parameter 'test'"):
        SchemaValidator(
            schema=cs.arguments_v3_schema(
                [
                    cs.arguments_v3_parameter(name='test', schema=cs.int_schema()),
                    cs.arguments_v3_parameter(name='a', schema=cs.int_schema()),
                    cs.arguments_v3_parameter(name='test', schema=cs.int_schema()),
                ]
            )
        )


def test_invalid_positional_only_parameter_position() -> None:
    with pytest.raises(SchemaError, match="Positional only parameter 'test' cannot follow other parameter kinds"):
        SchemaValidator(
            schema=cs.arguments_v3_schema(
                [
                    cs.arguments_v3_parameter(name='a', schema=cs.int_schema(), mode='var_args'),
                    cs.arguments_v3_parameter(name='test', schema=cs.int_schema(), mode='positional_only'),
                ]
            )
        )


def test_invalid_positional_or_keyword_parameter_position() -> None:
    with pytest.raises(
        SchemaError, match="Positional or keyword parameter 'test' cannot follow variadic or keyword only parameters"
    ):
        SchemaValidator(
            schema=cs.arguments_v3_schema(
                [
                    cs.arguments_v3_parameter(name='a', schema=cs.int_schema(), mode='var_args'),
                    cs.arguments_v3_parameter(name='test', schema=cs.int_schema(), mode='positional_or_keyword'),
                ]
            )
        )


def test_invalid_var_args_parameter_position() -> None:
    with pytest.raises(
        SchemaError, match="Variadic positional parameter 'test' cannot follow variadic or keyword only parameters"
    ):
        SchemaValidator(
            schema=cs.arguments_v3_schema(
                [
                    cs.arguments_v3_parameter(name='a', schema=cs.int_schema(), mode='keyword_only'),
                    cs.arguments_v3_parameter(name='test', schema=cs.int_schema(), mode='var_args'),
                ]
            )
        )


def test_invalid_keyword_only_parameter_position() -> None:
    with pytest.raises(
        SchemaError, match="Keyword only parameter 'test' cannot follow variadic keyword only parameter"
    ):
        SchemaValidator(
            schema=cs.arguments_v3_schema(
                [
                    cs.arguments_v3_parameter(name='a', schema=cs.int_schema(), mode='var_kwargs_uniform'),
                    cs.arguments_v3_parameter(name='test', schema=cs.int_schema(), mode='keyword_only'),
                ]
            )
        )



import copy
import functools
import pickle
import sys
import unittest
from datetime import datetime, timedelta, timezone, tzinfo
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

from pydantic_core import SchemaValidator, TzInfo, core_schema


class _ALWAYS_EQ:
    """
    Object that is equal to anything.
    """

    def __eq__(self, other):
        return True

    def __ne__(self, other):
        return False


ALWAYS_EQ = _ALWAYS_EQ()


@functools.total_ordering
class _LARGEST:
    """
    Object that is greater than anything (except itself).
    """

    def __eq__(self, other):
        return isinstance(other, _LARGEST)

    def __lt__(self, other):
        return False


LARGEST = _LARGEST()


@functools.total_ordering
class _SMALLEST:
    """
    Object that is less than anything (except itself).
    """

    def __eq__(self, other):
        return isinstance(other, _SMALLEST)

    def __gt__(self, other):
        return False


SMALLEST = _SMALLEST()


pickle_choices = [(pickle, pickle, proto) for proto in range(pickle.HIGHEST_PROTOCOL + 1)]

HOUR = timedelta(hours=1).total_seconds()
ZERO = timedelta(0).total_seconds()


def first_sunday_on_or_after(dt):
    days_to_go = 6 - dt.weekday()
    if days_to_go:
        dt += timedelta(days_to_go)
    return dt


DSTSTART = datetime(1, 4, 1, 2)
DSTEND = datetime(1, 10, 25, 1)


class TestTzInfo(unittest.TestCase):
    """Adapted from CPython `timezone` tests

    Original tests are located here https://github.com/python/cpython/blob/a0bb4a39d1ca10e4a75f50a9fbe90cc9db28d29e/Lib/test/datetimetester.py#L256
    """

    def setUp(self):
        self.ACDT = TzInfo(timedelta(hours=9.5).total_seconds())
        self.EST = TzInfo(-timedelta(hours=5).total_seconds())
        self.UTC = TzInfo(timedelta(0).total_seconds())
        self.DT = datetime(2010, 1, 1)

    def test_str(self):
        for tz in [self.ACDT, self.EST]:
            self.assertEqual(str(tz), tz.tzname(None))

    def test_constructor(self):
        for subminute in [timedelta(microseconds=1), timedelta(seconds=1)]:
            tz = TzInfo(subminute.total_seconds())
            self.assertNotEqual(tz.utcoffset(None) % timedelta(minutes=1), 0)
        # invalid offsets
        for invalid in [timedelta(1, 1), timedelta(1)]:
            self.assertRaises(ValueError, TzInfo, invalid.total_seconds())
            self.assertRaises(ValueError, TzInfo, -invalid.total_seconds())

        with self.assertRaises(TypeError):
            TzInfo(None)
        with self.assertRaises(TypeError):
            TzInfo(timedelta(seconds=42))
        with self.assertRaises(TypeError):
            TzInfo(ZERO, None)
        with self.assertRaises(TypeError):
            TzInfo(ZERO, 42)
        with self.assertRaises(TypeError):
            TzInfo(ZERO, 'ABC', 'extra')

    def test_inheritance(self):
        self.assertIsInstance(self.EST, tzinfo)

    def test_utcoffset(self):
        dummy = self.DT
        for h in [0, 1.5, 12]:
            offset = h * HOUR
            self.assertEqual(timedelta(seconds=offset), TzInfo(offset).utcoffset(dummy))
            self.assertEqual(timedelta(seconds=-offset), TzInfo(-offset).utcoffset(dummy))

        self.assertEqual(self.EST.utcoffset(''), timedelta(hours=-5))
        self.assertEqual(self.EST.utcoffset(5), timedelta(hours=-5))

    def test_dst(self):
        self.EST.dst('') is None
        self.EST.dst(5) is None

    def test_tzname(self):
        self.assertEqual('-05:00', TzInfo(-5 * HOUR).tzname(None))
        self.assertEqual('+09:30', TzInfo(9.5 * HOUR).tzname(None))
        self.assertEqual('-00:01', TzInfo(timedelta(minutes=-1).total_seconds()).tzname(None))
        # Sub-minute offsets:
        self.assertEqual('+01:06:40', TzInfo(timedelta(0, 4000).total_seconds()).tzname(None))
        self.assertEqual('-01:06:40', TzInfo(-timedelta(0, 4000).total_seconds()).tzname(None))
        self.assertEqual('+01:06:40', TzInfo(timedelta(0, 4000, 1).total_seconds()).tzname(None))
        self.assertEqual('-01:06:40', TzInfo(-timedelta(0, 4000, 1).total_seconds()).tzname(None))

        self.assertEqual(self.EST.tzname(''), '-05:00')
        self.assertEqual(self.EST.tzname(5), '-05:00')

    def test_fromutc(self):
        for tz in [self.EST, self.ACDT]:
            utctime = self.DT.replace(tzinfo=tz)
            local = tz.fromutc(utctime)
            self.assertEqual(local - utctime, tz.utcoffset(local))
            self.assertEqual(local, self.DT.replace(tzinfo=timezone.utc))

    def test_comparison(self):
        self.assertNotEqual(TzInfo(ZERO), TzInfo(HOUR))
        self.assertEqual(TzInfo(HOUR), TzInfo(HOUR))
        self.assertFalse(TzInfo(ZERO) < TzInfo(ZERO))
        self.assertIn(TzInfo(ZERO), {TzInfo(ZERO)})
        self.assertTrue(TzInfo(ZERO) is not None)
        self.assertFalse(TzInfo(ZERO) is None)

        tz = TzInfo(ZERO)
        self.assertTrue(tz == ALWAYS_EQ)
        self.assertFalse(tz != ALWAYS_EQ)
        self.assertTrue(tz < LARGEST)
        self.assertFalse(tz > LARGEST)
        self.assertTrue(tz <= LARGEST)
        self.assertFalse(tz >= LARGEST)
        self.assertFalse(tz < SMALLEST)
        self.assertTrue(tz > SMALLEST)
        self.assertFalse(tz <= SMALLEST)
        self.assertTrue(tz >= SMALLEST)

        # offset based comparion tests for tzinfo derived classes like datetime.timezone.
        utcdatetime = self.DT.replace(tzinfo=timezone.utc)
        self.assertTrue(tz == utcdatetime.tzinfo)
        estdatetime = self.DT.replace(tzinfo=timezone(-timedelta(hours=5)))
        self.assertTrue(self.EST == estdatetime.tzinfo)
        self.assertTrue(tz > estdatetime.tzinfo)

        if sys.platform == 'linux':
            try:
                europe_london = ZoneInfo('Europe/London')
            except ZoneInfoNotFoundError:
                # tz data not available
                pass
            else:
                self.assertFalse(tz == europe_london)
                with self.assertRaises(TypeError):
                    tz > europe_london

    def test_copy(self):
        for tz in self.ACDT, self.EST:
            tz_copy = copy.copy(tz)
            self.assertEqual(tz_copy, tz)

    def test_deepcopy(self):
        for tz in self.ACDT, self.EST:
            tz_copy = copy.deepcopy(tz)
            self.assertEqual(tz_copy, tz)

    def test_offset_boundaries(self):
        # Test timedeltas close to the boundaries
        time_deltas = [timedelta(hours=23, minutes=59), timedelta(hours=23, minutes=59, seconds=59)]
        time_deltas.extend([-delta for delta in time_deltas])

        for delta in time_deltas:
            with self.subTest(test_type='good', delta=delta):
                print(delta.total_seconds())
                TzInfo(delta.total_seconds())

        # Test timedeltas on and outside the boundaries
        bad_time_deltas = [timedelta(hours=24), timedelta(hours=24, microseconds=1)]
        bad_time_deltas.extend([-delta for delta in bad_time_deltas])

        for delta in bad_time_deltas:
            with self.subTest(test_type='bad', delta=delta):
                with self.assertRaises(ValueError):
                    TzInfo(delta.total_seconds())

    def test_no_args_constructor(self):
        # Test that TzInfo can be constructed without arguments
        tz = TzInfo()
        self.assertEqual(tz.utcoffset(None), timedelta(0))
        self.assertEqual(str(tz), 'UTC')

    def test_pickle(self):
        # Test that TzInfo can be pickled and unpickled
        for tz in self.ACDT, self.EST, self.UTC:
            for pickler, unpickler, proto in pickle_choices:
                with self.subTest(tz=tz, proto=proto):
                    pickled = pickler.dumps(tz, proto)
                    unpickled = unpickler.loads(pickled)
                    self.assertEqual(tz, unpickled)
                    self.assertEqual(tz.utcoffset(None), unpickled.utcoffset(None))


def test_tzinfo_could_be_reused():
    class Model:
        value: datetime

    v = SchemaValidator(
        core_schema.model_schema(
            Model, core_schema.model_fields_schema({'value': core_schema.model_field(core_schema.datetime_schema())})
        )
    )

    m = v.validate_python({'value': '2015-10-21T15:28:00.000000+01:00'})

    target = datetime(1955, 11, 12, 14, 38, tzinfo=m.value.tzinfo)
    assert target == datetime(1955, 11, 12, 14, 38, tzinfo=timezone(timedelta(hours=1)))

    now = datetime.now(tz=m.value.tzinfo)
    assert isinstance(now, datetime)



import pickle

import pytest

from pydantic_core import SchemaValidator
from pydantic_core import core_schema as cs


def test_schema_as_string():
    v = SchemaValidator(cs.bool_schema())
    assert v.validate_python('tRuE') is True


@pytest.mark.parametrize('pickle_protocol', range(1, pickle.HIGHEST_PROTOCOL + 1))
def test_pickle(pickle_protocol: int) -> None:
    v1 = SchemaValidator(cs.bool_schema())
    assert v1.validate_python('tRuE') is True
    p = pickle.dumps(v1, protocol=pickle_protocol)
    v2 = pickle.loads(p)
    assert v2.validate_python('tRuE') is True
    assert repr(v1) == repr(v2)


def test_not_schema_definition_error():
    schema = {
        'type': 'typed-dict',
        'fields': {
            f'f_{i}': {'type': 'typed-dict-field', 'schema': {'type': 'nullable', 'schema': {'type': 'int'}}}
            for i in range(101)
        },
    }
    v = SchemaValidator(schema)
    assert repr(v).count('TypedDictField') == 101


def test_try_self_schema_discriminator():
    """Trying to use self-schema when it shouldn't be used"""
    v = SchemaValidator(cs.tagged_union_schema(choices={'int': cs.int_schema()}, discriminator='self-schema'))
    assert 'discriminator: LookupKey' in repr(v)


def test_build_recursive_schema_from_defs() -> None:
    """
    Validate a schema representing mutually recursive models, analogous to the following JSON schema:

    ```json
    {
        "$schema": "https://json-schema.org/draft/2019-09/schema",
        "oneOf": [{"$ref": "#/$defs/a"}],
        "$defs": {
            "a": {
                "type": "object",
                "properties": {"b": {"type": "array", "items": {"$ref": "#/$defs/a"}}},
                "required": ["b"],
            },
            "b": {
                "type": "object",
                "properties": {"a": {"type": "array", "items": {"$ref": "#/$defs/b"}}},
                "required": ["a"],
            },
        },
    }
    ```
    """

    s = cs.definitions_schema(
        cs.definition_reference_schema(schema_ref='a'),
        [
            cs.typed_dict_schema(
                {'b': cs.typed_dict_field(cs.list_schema(cs.definition_reference_schema('b')))}, ref='a'
            ),
            cs.typed_dict_schema(
                {'a': cs.typed_dict_field(cs.list_schema(cs.definition_reference_schema('a')))}, ref='b'
            ),
        ],
    )

    SchemaValidator(s)



import platform
import re
import sys
from decimal import Decimal
from numbers import Number
from typing import Any, Union

import pytest

from pydantic_core import CoreConfig, SchemaError, SchemaValidator, ValidationError, core_schema

from ..conftest import Err, PyAndJson, plain_repr


@pytest.mark.parametrize(
    'input_value,expected',
    [
        ('foobar', 'foobar'),
        (123, Err('Input should be a valid string [type=string_type, input_value=123, input_type=int]')),
        (123.456, Err('Input should be a valid string [type=string_type, input_value=123.456, input_type=float]')),
        (False, Err('Input should be a valid string [type=string_type')),
        (True, Err('Input should be a valid string [type=string_type')),
        ([], Err('Input should be a valid string [type=string_type, input_value=[], input_type=list]')),
    ],
)
def test_str(py_and_json: PyAndJson, input_value, expected):
    v = py_and_json({'type': 'str'})
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_test(input_value)
    else:
        assert v.validate_test(input_value) == expected


@pytest.mark.parametrize(
    'input_value,expected',
    [
        ('foobar', 'foobar'),
        ('🐈 Hello \ud800World', '🐈 Hello \ud800World'),
        (b'foobar', 'foobar'),
        (bytearray(b'foobar'), 'foobar'),
        (
            b'\x81',
            Err('Input should be a valid string, unable to parse raw data as a unicode string [type=string_unicode'),
        ),
        (
            bytearray(b'\x81'),
            Err('Input should be a valid string, unable to parse raw data as a unicode string [type=string_unicode'),
        ),
        # null bytes are very annoying, but we can't really block them here
        (b'\x00', '\x00'),
        (123, Err('Input should be a valid string [type=string_type, input_value=123, input_type=int]')),
        (
            Decimal('123'),
            Err("Input should be a valid string [type=string_type, input_value=Decimal('123'), input_type=Decimal]"),
        ),
    ],
)
def test_str_not_json(input_value, expected):
    v = SchemaValidator(core_schema.str_schema())
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_python(input_value)
    else:
        assert v.validate_python(input_value) == expected


@pytest.mark.parametrize(
    'kwargs,input_value,expected',
    [
        ({}, 'abc', 'abc'),
        ({'strict': True}, 'Foobar', 'Foobar'),
        ({'to_upper': True}, 'fooBar', 'FOOBAR'),
        ({'to_lower': True}, 'fooBar', 'foobar'),
        ({'strip_whitespace': True}, ' foobar  ', 'foobar'),
        ({'strip_whitespace': True, 'to_upper': True}, ' fooBar', 'FOOBAR'),
        ({'min_length': 5}, '12345', '12345'),
        ({'min_length': 5}, '1234', Err('String should have at least 5 characters [type=string_too_short')),
        ({'max_length': 5}, '12345', '12345'),
        ({'max_length': 5}, '123456', Err('String should have at most 5 characters [type=string_too_long')),
        ({'pattern': r'^\d+$'}, '12345', '12345'),
        ({'pattern': r'\d+$'}, 'foobar 123', 'foobar 123'),
        ({'pattern': r'^\d+$'}, '12345a', Err("String should match pattern '^\\d+$' [type=string_pattern_mismatch")),
        ({'pattern': r'[a-z]'}, 'Abc', 'Abc'),
        ({'pattern': re.compile(r'[a-z]')}, 'Abc', 'Abc'),
        # strip comes after length check
        ({'max_length': 5, 'strip_whitespace': True}, '1234  ', '1234'),
        # to_upper and strip comes after pattern check
        ({'to_upper': True, 'pattern': 'abc'}, 'abc', 'ABC'),
        ({'strip_whitespace': True, 'pattern': r'\d+$'}, 'foobar 123 ', 'foobar 123'),
        ({'min_length': 1}, '🐈 Hello', '🐈 Hello'),
    ],
)
def test_constrained_str(py_and_json: PyAndJson, kwargs: dict[str, Any], input_value, expected):
    v = py_and_json({'type': 'str', **kwargs})
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_test(input_value)
    else:
        assert v.validate_test(input_value) == expected


@pytest.mark.parametrize(
    'kwargs,input_value,expected',
    [
        ({}, b'abc', 'abc'),
        ({'strict': True}, 'Foobar', 'Foobar'),
        (
            {'strict': True},
            123,
            Err('Input should be a valid string [type=string_type, input_value=123, input_type=int]'),
        ),
    ],
)
def test_constrained_str_py_only(kwargs: dict[str, Any], input_value, expected):
    v = SchemaValidator(core_schema.str_schema(**kwargs))
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_python(input_value)
    else:
        assert v.validate_python(input_value) == expected


def test_unicode_error():
    # `.to_str()` Returns a `UnicodeEncodeError` if the input is not valid unicode (containing unpaired surrogates).
    # https://github.com/PyO3/pyo3/blob/6503128442b8f3e767c663a6a8d96376d7fb603d/src/types/string.rs#L477
    v = SchemaValidator(core_schema.str_schema(min_length=1))
    assert v.validate_python('🐈 Hello') == '🐈 Hello'

    with pytest.raises(ValidationError) as exc_info:
        v.validate_python('🐈 Hello \ud800World')
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'string_unicode',
            'loc': (),
            'msg': 'Input should be a valid string, unable to parse raw data as a unicode string',
            'input': '🐈 Hello \ud800World',
        }
    ]


@pytest.mark.parametrize(
    ('data', 'max_length', 'error'),
    [
        pytest.param('test', 5, None, id='short string'),
        pytest.param('test long', 5, 'String should have at most 5 characters', id='long string'),
        pytest.param('␛⯋℃▤', 5, None, id='short string with unicode characters'),
        pytest.param(
            '␛⯋℃▤⩥⠫⳼⣪⨺✒⧐♳⩚⏭⏣⍥┙⧃Ⰴ┽⏏♜',
            5,
            'String should have at most 5 characters',
            id='long string with unicode characters',
        ),
        pytest.param('а' * 25, 32, None, id='a lot of `а`s'),
    ],
)
def test_str_constrained(data: str, max_length: int, error: Union[re.Pattern, None]):
    v = SchemaValidator(core_schema.str_schema(max_length=max_length))
    if error is None:
        assert v.validate_python(data) == data
    else:
        with pytest.raises(ValidationError, match=error):
            v.validate_python(data)


def test_str_constrained_config():
    v = SchemaValidator(core_schema.str_schema(), config=CoreConfig(str_max_length=5))
    assert v.validate_python('test') == 'test'

    with pytest.raises(ValidationError, match='String should have at most 5 characters'):
        v.validate_python('test long')


@pytest.mark.parametrize('engine', [None, 'rust-regex', 'python-re'])
def test_invalid_regex(engine):
    if platform.python_implementation() == 'PyPy' and sys.version_info[:2] == (3, 11):
        # pypy 3.11 type formatting
        pytest.xfail()

    # TODO uncomment and fix once #150 is done
    # with pytest.raises(SchemaError) as exc_info:
    #     SchemaValidator({'type': 'str', 'pattern': 123})
    # assert exc_info.value.args[0] == (
    #     'Error building "str" validator:\n  TypeError: \'int\' object cannot be converted to \'PyString\''
    # )
    with pytest.raises(SchemaError) as exc_info:
        SchemaValidator(core_schema.str_schema(pattern='(abc', regex_engine=engine))

    if engine is None or engine == 'rust-regex':
        assert exc_info.value.args[0] == (
            'Error building "str" validator:\n  SchemaError: regex parse error:\n    (abc\n    ^\nerror: unclosed group'
        )
    elif engine == 'python-re':
        prefix = 'PatternError' if sys.version_info >= (3, 13) else 'error'
        assert exc_info.value.args[0] == (
            f'Error building "str" validator:\n  {prefix}: missing ), unterminated subpattern at position 0'
        )


@pytest.mark.parametrize('engine', [None, 'rust-regex', 'python-re'])
def test_regex_error(engine):
    v = SchemaValidator(core_schema.str_schema(pattern='11', regex_engine=engine))
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python('12')
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'string_pattern_mismatch',
            'loc': (),
            'msg': "String should match pattern '11'",
            'input': '12',
            'ctx': {'pattern': '11'},
        }
    ]


def test_default_validator():
    v = SchemaValidator(
        core_schema.str_schema(strict=True, to_lower=False), config=CoreConfig(str_strip_whitespace=False)
    )
    assert (
        plain_repr(v)
        == 'SchemaValidator(title="str",validator=Str(StrValidator{strict:true,coerce_numbers_to_str:false}),definitions=[],cache_strings=True)'
    )


@pytest.fixture(scope='session', name='FruitEnum')
def fruit_enum_fixture():
    from enum import Enum

    class FruitEnum(str, Enum):
        pear = 'pear'
        banana = 'banana'

    return FruitEnum


@pytest.mark.parametrize('to_lower', [False, True], ids=repr)
def test_strict_subclass(to_lower: bool):
    v = SchemaValidator(core_schema.str_schema(strict=True, to_lower=to_lower))

    class StrSubclass(str):
        pass

    res = v.validate_python(StrSubclass('ABC'))
    assert res == 'abc' if to_lower else 'ABC'


@pytest.mark.parametrize('kwargs', [{}, {'to_lower': True}], ids=repr)
def test_lax_subclass(FruitEnum, kwargs):
    v = SchemaValidator(core_schema.str_schema(**kwargs))
    assert v.validate_python('foobar') == 'foobar'
    assert v.validate_python(b'foobar') == 'foobar'
    p = v.validate_python(FruitEnum.pear)
    assert p == 'pear'
    assert type(p) is str
    assert repr(p) == "'pear'"


@pytest.mark.parametrize('kwargs', [{}, {'to_lower': True}], ids=repr)
def test_lax_subclass_plain_enum(kwargs):
    v = SchemaValidator(core_schema.str_schema(**kwargs))

    from enum import Enum

    class PlainEnum(Enum):
        ONE = 'one'

    p = v.validate_python(PlainEnum.ONE)
    assert p == 'one'
    assert type(p) is str
    assert repr(p) == "'one'"


def test_subclass_preserved() -> None:
    class StrSubclass(str):
        pass

    v = SchemaValidator(core_schema.str_schema())

    assert not isinstance(v.validate_python(StrSubclass('')), StrSubclass)
    assert not isinstance(v.validate_python(StrSubclass(''), strict=True), StrSubclass)

    # unions do a first pass in strict mode
    # so verify that they don't match the str schema in strict mode
    # and preserve the type
    v = SchemaValidator(core_schema.union_schema([core_schema.str_schema(), core_schema.int_schema()]))

    assert not isinstance(v.validate_python(StrSubclass('')), StrSubclass)
    assert not isinstance(v.validate_python(StrSubclass(''), strict=True), StrSubclass)


@pytest.mark.parametrize('string', [True, False])
def test_coerce_numbers_to_str_with_invalid_unicode_character(string) -> None:
    config = core_schema.CoreConfig(coerce_numbers_to_str=True)

    v = SchemaValidator(core_schema.str_schema(strict=string), config=config)
    assert v.validate_python('\ud835') == '\ud835'


def test_coerce_numbers_to_str_disabled_in_strict_mode() -> None:
    config = core_schema.CoreConfig(coerce_numbers_to_str=True)

    v = SchemaValidator(core_schema.str_schema(strict=True), config=config)
    with pytest.raises(ValidationError):
        v.validate_python(42)
    with pytest.raises(ValidationError):
        v.validate_json('42')


def test_coerce_numbers_to_str_raises_for_bool() -> None:
    config = core_schema.CoreConfig(coerce_numbers_to_str=True)

    v = SchemaValidator(core_schema.str_schema(), config=config)
    with pytest.raises(ValidationError):
        v.validate_python(True)
    with pytest.raises(ValidationError):
        v.validate_json(False)


@pytest.mark.parametrize(
    ('number', 'expected_str'),
    [
        pytest.param(42, '42', id='42'),
        pytest.param(42.0, '42.0', id='42.0'),
        pytest.param(Decimal('42.0'), '42.0', id="Decimal('42.0')"),
    ],
)
def test_coerce_numbers_to_str(number: Number, expected_str: str) -> None:
    config = core_schema.CoreConfig(coerce_numbers_to_str=True)

    v = SchemaValidator(core_schema.str_schema(), config=config)
    assert v.validate_python(number) == expected_str


@pytest.mark.parametrize(
    ('number', 'expected_str'),
    [
        pytest.param('42', '42', id='42'),
        pytest.param('42.0', '42', id='42.0'),
        pytest.param('42.13', '42.13', id='42.13'),
    ],
)
def test_coerce_numbers_to_str_from_json(number: str, expected_str: str) -> None:
    config = core_schema.CoreConfig(coerce_numbers_to_str=True)

    v = SchemaValidator(core_schema.str_schema(), config=config)
    assert v.validate_json(number) == expected_str


@pytest.mark.parametrize('mode', (None, 'schema', 'config'))
@pytest.mark.xfail(
    platform.python_implementation() == 'PyPy' and sys.version_info[:2] == (3, 11), reason='pypy 3.11 type formatting'
)
def test_backtracking_regex_rust_unsupported(mode) -> None:
    pattern = r'r(#*)".*?"\1'

    with pytest.raises(SchemaError) as exc_info:
        if mode is None:
            # rust-regex is the default
            SchemaValidator(core_schema.str_schema(pattern=pattern))
        elif mode == 'schema':
            SchemaValidator(core_schema.str_schema(pattern=pattern, regex_engine='rust-regex'))
        elif mode == 'config':
            SchemaValidator(
                schema=core_schema.str_schema(pattern=pattern), config=core_schema.CoreConfig(regex_engine='rust-regex')
            )

    assert exc_info.value.args[0] == (
        'Error building "str" validator:\n'
        '  SchemaError: regex parse error:\n'
        '    r(#*)".*?"\\1\n'
        '              ^^\n'
        'error: backreferences are not supported'
    )


@pytest.mark.parametrize('mode', ('schema', 'config'))
def test_backtracking_regex_python(mode) -> None:
    pattern = r'r(#*)".*?"\1'

    if mode == 'schema':
        v = SchemaValidator(core_schema.str_schema(pattern=pattern, regex_engine='python-re'))
    elif mode == 'config':
        v = SchemaValidator(
            schema=core_schema.str_schema(pattern=pattern), config=core_schema.CoreConfig(regex_engine='python-re')
        )
    assert v.validate_python('r""') == 'r""'
    assert v.validate_python('r#""#') == 'r#""#'
    with pytest.raises(ValidationError):
        # not a valid match for the pattern
        v.validate_python('r#"#')


@pytest.mark.parametrize('number', (42, 443, 10242))
def test_coerce_numbers_to_str_schema(number: int):
    v = SchemaValidator(core_schema.str_schema(coerce_numbers_to_str=True))
    assert v.validate_python(number) == str(number)
    assert v.validate_json(str(number)) == str(number)


@pytest.mark.parametrize('number', (42, 443, 10242))
def test_coerce_numbers_to_str_schema_precedence(number: int):
    config = core_schema.CoreConfig(coerce_numbers_to_str=False)
    v = SchemaValidator(core_schema.str_schema(coerce_numbers_to_str=True), config=config)
    assert v.validate_python(number) == str(number)
    assert v.validate_json(str(number)) == str(number)

    config = core_schema.CoreConfig(coerce_numbers_to_str=True)
    v = SchemaValidator(core_schema.str_schema(coerce_numbers_to_str=False), config=config)
    with pytest.raises(ValidationError):
        v.validate_python(number)
    with pytest.raises(ValidationError):
        v.validate_json(str(number))


@pytest.mark.parametrize('number', (42, 443, 10242))
def test_coerce_numbers_to_str_schema_with_strict_mode(number: int):
    v = SchemaValidator(core_schema.str_schema(coerce_numbers_to_str=True, strict=True))
    with pytest.raises(ValidationError):
        v.validate_python(number)
    with pytest.raises(ValidationError):
        v.validate_json(str(number))


@pytest.mark.parametrize('engine', [None, 'rust-regex', 'python-re'])
def test_compiled_regex(engine) -> None:
    v = SchemaValidator(core_schema.str_schema(pattern=re.compile('abc', re.IGNORECASE), regex_engine=engine))
    assert v.validate_python('abc') == 'abc'
    assert v.validate_python('ABC') == 'ABC'



import json

import pytest
from dirty_equals import IsList

from pydantic_core import SchemaSerializer, core_schema


def test_set_any():
    v = SchemaSerializer(core_schema.set_schema(core_schema.any_schema()))
    assert v.to_python({'a', 'b', 'c'}) == {'a', 'b', 'c'}
    assert v.to_python({'a', 'b', 'c'}, mode='json') == IsList('a', 'b', 'c', check_order=False)
    assert json.loads(v.to_json({'a', 'b', 'c'})) == IsList('a', 'b', 'c', check_order=False)


def test_frozenset_any():
    v = SchemaSerializer(core_schema.frozenset_schema(core_schema.any_schema()))
    fs = frozenset(['a', 'b', 'c'])
    output = v.to_python(fs)
    assert output == {'a', 'b', 'c'}
    assert type(output) == frozenset
    assert v.to_python(fs, mode='json') == IsList('a', 'b', 'c', check_order=False)
    assert json.loads(v.to_json(fs)) == IsList('a', 'b', 'c', check_order=False)


@pytest.mark.parametrize(
    'input_value,json_output,expected_type',
    [
        ('apple', 'apple', r'set\[int\]'),
        ([1, 2, 3], [1, 2, 3], r'set\[int\]'),
        ((1, 2, 3), [1, 2, 3], r'set\[int\]'),
        (
            frozenset([1, 2, 3]),
            IsList(1, 2, 3, check_order=False),
            r'set\[int\]',
        ),
        ({1, 2, 'a'}, IsList(1, 2, 'a', check_order=False), 'int'),
    ],
)
def test_set_fallback(input_value, json_output, expected_type):
    v = SchemaSerializer(core_schema.set_schema(core_schema.int_schema()))
    assert v.to_python({1, 2, 3}) == {1, 2, 3}

    with pytest.warns(
        UserWarning,
        match=f'Expected `{expected_type}` - serialized value may not be as expected',
    ):
        assert v.to_python(input_value) == input_value

    with pytest.warns(
        UserWarning,
        match=f'Expected `{expected_type}` - serialized value may not be as expected',
    ):
        assert v.to_python(input_value, mode='json') == json_output

    with pytest.warns(
        UserWarning,
        match=f'Expected `{expected_type}` - serialized value may not be as expected',
    ):
        assert json.loads(v.to_json(input_value)) == json_output



from typing import Any, Optional
from unittest import TestCase
from unittest.mock import ANY

import pytest
from typing_extensions import LiteralString, Self, override

from pydantic_core import ErrorDetails, InitErrorDetails, PydanticCustomError, ValidationError


def test_validation_error_subclassable():
    """Assert subclassable and inheritance hierarchy as expected"""

    class CustomValidationError(ValidationError):
        pass

    with pytest.raises(ValidationError) as exception_info:
        raise CustomValidationError.from_exception_data(
            'My CustomError',
            [
                InitErrorDetails(
                    type='value_error',
                    loc=('myField',),
                    msg='This is my custom error.',
                    input='something invalid',
                    ctx={
                        'myField': 'something invalid',
                        'error': "'something invalid' is not a valid value for 'myField'",
                    },
                )
            ],
        )
    assert isinstance(exception_info.value, CustomValidationError)


def test_validation_error_loc_overrides():
    """Override methods in rust pyclass and assert change in behavior: ValidationError.errors"""

    class CustomLocOverridesError(ValidationError):
        """Unnests some errors"""

        @override
        def errors(
            self, *, include_url: bool = True, include_context: bool = True, include_input: bool = True
        ) -> list[ErrorDetails]:
            errors = super().errors(
                include_url=include_url, include_context=include_context, include_input=include_input
            )
            return [{**error, 'loc': error['loc'][1:]} for error in errors]

    with pytest.raises(CustomLocOverridesError) as exception_info:
        raise CustomLocOverridesError.from_exception_data(
            'My CustomError',
            [
                InitErrorDetails(
                    type='value_error',
                    loc=(
                        'hide_this',
                        'myField',
                    ),
                    msg='This is my custom error.',
                    input='something invalid',
                    ctx={
                        'myField': 'something invalid',
                        'error': "'something invalid' is not a valid value for 'myField'",
                    },
                ),
                InitErrorDetails(
                    type='value_error',
                    loc=(
                        'hide_this',
                        'myFieldToo',
                    ),
                    msg='This is my custom error.',
                    input='something invalid',
                    ctx={
                        'myFieldToo': 'something invalid',
                        'error': "'something invalid' is not a valid value for 'myFieldToo'",
                    },
                ),
            ],
        )

    TestCase().assertCountEqual(
        exception_info.value.errors(),
        [
            {
                'type': 'value_error',
                'loc': ('myField',),
                'msg': "Value error, 'something invalid' is not a valid value for 'myField'",
                'input': 'something invalid',
                'ctx': {
                    'error': "'something invalid' is not a valid value for 'myField'",
                    'myField': 'something invalid',
                },
                'url': ANY,
            },
            {
                'type': 'value_error',
                'loc': ('myFieldToo',),
                'msg': "Value error, 'something invalid' is not a valid value for 'myFieldToo'",
                'input': 'something invalid',
                'ctx': {
                    'error': "'something invalid' is not a valid value for 'myFieldToo'",
                    'myFieldToo': 'something invalid',
                },
                'url': ANY,
            },
        ],
    )


def test_custom_pydantic_error_subclassable():
    """Assert subclassable and inheritance hierarchy as expected"""

    class MyCustomError(PydanticCustomError):
        pass

    with pytest.raises(PydanticCustomError) as exception_info:
        raise MyCustomError(
            'not_my_custom_thing',
            "value is not compatible with my custom field, got '{wrong_value}'",
            {'wrong_value': 'non compatible value'},
        )
    assert isinstance(exception_info.value, MyCustomError)


def test_custom_pydantic_error_overrides():
    """Override methods in rust pyclass and assert change in behavior: PydanticCustomError.__new__"""

    class CustomErrorWithCustomTemplate(PydanticCustomError):
        @override
        def __new__(
            cls, error_type: LiteralString, my_custom_setting: str, context: Optional[dict[str, Any]] = None
        ) -> Self:
            message_template = (
                "'{my_custom_value}' setting requires a specific my custom field value, got '{wrong_value}'"
            )
            context = {**context, 'my_custom_value': my_custom_setting}
            return super().__new__(cls, error_type, message_template, context)

    with pytest.raises(CustomErrorWithCustomTemplate) as exception_info:
        raise CustomErrorWithCustomTemplate(
            'not_my_custom_thing', 'my_setting', {'wrong_value': 'non compatible value'}
        )

    assert (
        exception_info.value.message()
        == "'my_setting' setting requires a specific my custom field value, got 'non compatible value'"
    )



from dataclasses import dataclass
from typing import Optional

from typing_extensions import TypedDict

from pydantic_core import SchemaSerializer, SchemaValidator, core_schema


def test_serialize_as_any_with_models() -> None:
    class Parent:
        x: int

    class Child(Parent):
        y: str

    Parent.__pydantic_core_schema__ = core_schema.model_schema(
        Parent,
        core_schema.model_fields_schema(
            {
                'x': core_schema.model_field(core_schema.int_schema()),
            }
        ),
    )
    Parent.__pydantic_validator__ = SchemaValidator(Parent.__pydantic_core_schema__)
    Parent.__pydantic_serializer__ = SchemaSerializer(Parent.__pydantic_core_schema__)

    Child.__pydantic_core_schema__ = core_schema.model_schema(
        Child,
        core_schema.model_fields_schema(
            {
                'x': core_schema.model_field(core_schema.int_schema()),
                'y': core_schema.model_field(core_schema.str_schema()),
            }
        ),
    )
    Child.__pydantic_validator__ = SchemaValidator(Child.__pydantic_core_schema__)
    Child.__pydantic_serializer__ = SchemaSerializer(Child.__pydantic_core_schema__)

    child = Child.__pydantic_validator__.validate_python({'x': 1, 'y': 'hopefully not a secret'})
    assert Parent.__pydantic_serializer__.to_python(child, serialize_as_any=False) == {'x': 1}
    assert Parent.__pydantic_serializer__.to_python(child, serialize_as_any=True) == {
        'x': 1,
        'y': 'hopefully not a secret',
    }


def test_serialize_as_any_with_dataclass() -> None:
    @dataclass
    class Parent:
        x: int

    class Child(Parent):
        y: str

    Parent.__pydantic_core_schema__ = core_schema.dataclass_schema(
        Parent,
        core_schema.dataclass_args_schema(
            'Parent',
            [
                core_schema.dataclass_field(name='x', schema=core_schema.int_schema()),
            ],
        ),
        ['x'],
    )
    Parent.__pydantic_validator__ = SchemaValidator(Parent.__pydantic_core_schema__)
    Parent.__pydantic_serializer__ = SchemaSerializer(Parent.__pydantic_core_schema__)

    Child.__pydantic_core_schema__ = core_schema.dataclass_schema(
        Child,
        core_schema.dataclass_args_schema(
            'Child',
            [
                core_schema.dataclass_field(name='x', schema=core_schema.int_schema()),
                core_schema.dataclass_field(name='y', schema=core_schema.str_schema()),
            ],
        ),
        ['x', 'y'],
    )
    Child.__pydantic_validator__ = SchemaValidator(Child.__pydantic_core_schema__)
    Child.__pydantic_serializer__ = SchemaSerializer(Child.__pydantic_core_schema__)

    child = Child.__pydantic_validator__.validate_python({'x': 1, 'y': 'hopefully not a secret'})
    assert Parent.__pydantic_serializer__.to_python(child, serialize_as_any=False) == {'x': 1}
    assert Parent.__pydantic_serializer__.to_python(child, serialize_as_any=True) == {
        'x': 1,
        'y': 'hopefully not a secret',
    }


def test_serialize_as_any_with_typeddict() -> None:
    class Parent(TypedDict):
        x: int

    class Child(Parent):
        y: str

    Parent.__pydantic_core_schema__ = core_schema.typed_dict_schema(
        {
            'x': core_schema.typed_dict_field(core_schema.int_schema()),
        }
    )
    Parent.__pydantic_validator__ = SchemaValidator(Parent.__pydantic_core_schema__)
    Parent.__pydantic_serializer__ = SchemaSerializer(Parent.__pydantic_core_schema__)

    Child.__pydantic_core_schema__ = core_schema.typed_dict_schema(
        {
            'x': core_schema.typed_dict_field(core_schema.int_schema()),
            'y': core_schema.typed_dict_field(core_schema.str_schema()),
        }
    )
    Child.__pydantic_validator__ = SchemaValidator(Child.__pydantic_core_schema__)
    Child.__pydantic_serializer__ = SchemaSerializer(Child.__pydantic_core_schema__)

    child = Child.__pydantic_validator__.validate_python({'x': 1, 'y': 'hopefully not a secret'})
    assert Parent.__pydantic_serializer__.to_python(child, serialize_as_any=False) == {'x': 1}
    assert Parent.__pydantic_serializer__.to_python(child, serialize_as_any=True) == {
        'x': 1,
        'y': 'hopefully not a secret',
    }


def test_serialize_as_any_with_unrelated_models() -> None:
    class Parent:
        x: int

    class Other:
        y: str

    Parent.__pydantic_core_schema__ = core_schema.model_schema(
        Parent,
        core_schema.model_fields_schema(
            {
                'x': core_schema.model_field(core_schema.int_schema()),
            }
        ),
    )
    Parent.__pydantic_validator__ = SchemaValidator(Parent.__pydantic_core_schema__)
    Parent.__pydantic_serializer__ = SchemaSerializer(Parent.__pydantic_core_schema__)

    Other.__pydantic_core_schema__ = core_schema.model_schema(
        Other,
        core_schema.model_fields_schema(
            {
                'y': core_schema.model_field(core_schema.str_schema()),
            }
        ),
        config=core_schema.CoreConfig(extra_fields_behavior='allow'),
    )
    Other.__pydantic_validator__ = SchemaValidator(Other.__pydantic_core_schema__)
    Other.__pydantic_serializer__ = SchemaSerializer(Other.__pydantic_core_schema__)

    other = Other.__pydantic_validator__.validate_python({'x': 1, 'y': 'hopefully not a secret'})
    assert Parent.__pydantic_serializer__.to_python(other, serialize_as_any=False) == {}
    # note, without extra='allow', the 'x' field would not be included, as it's not in the schema
    assert Parent.__pydantic_serializer__.to_python(other, serialize_as_any=True) == {
        'x': 1,
        'y': 'hopefully not a secret',
    }


def test_serialize_as_any_with_nested_models() -> None:
    class Parent:
        x: int

    class Other(Parent):
        y: str

    class Outer:
        p: Parent

    Parent.__pydantic_core_schema__ = core_schema.model_schema(
        Parent,
        core_schema.model_fields_schema(
            {
                'x': core_schema.model_field(core_schema.int_schema()),
            }
        ),
        ref='Parent',
    )
    Parent.__pydantic_validator__ = SchemaValidator(Parent.__pydantic_core_schema__)
    Parent.__pydantic_serializer__ = SchemaSerializer(Parent.__pydantic_core_schema__)

    Other.__pydantic_core_schema__ = core_schema.model_schema(
        Other,
        core_schema.model_fields_schema(
            {
                'x': core_schema.model_field(core_schema.int_schema()),
                'y': core_schema.model_field(core_schema.str_schema()),
            }
        ),
        config=core_schema.CoreConfig(extra_fields_behavior='allow'),
    )
    Other.__pydantic_validator__ = SchemaValidator(Other.__pydantic_core_schema__)
    Other.__pydantic_serializer__ = SchemaSerializer(Other.__pydantic_core_schema__)

    Outer.__pydantic_core_schema__ = core_schema.definitions_schema(
        core_schema.model_schema(
            Outer,
            core_schema.model_fields_schema(
                {
                    'p': core_schema.model_field(core_schema.definition_reference_schema('Parent')),
                }
            ),
        ),
        [
            Parent.__pydantic_core_schema__,
        ],
    )
    Outer.__pydantic_validator__ = SchemaValidator(Outer.__pydantic_core_schema__)
    Outer.__pydantic_serializer__ = SchemaSerializer(Outer.__pydantic_core_schema__)

    other = Other.__pydantic_validator__.validate_python({'x': 1, 'y': 'hopefully not a secret'})
    outer = Outer()
    outer.p = other

    assert Outer.__pydantic_serializer__.to_python(outer, serialize_as_any=False) == {
        'p': {'x': 1},
    }
    assert Outer.__pydantic_serializer__.to_python(outer, serialize_as_any=True) == {
        'p': {
            'x': 1,
            'y': 'hopefully not a secret',
        }
    }

    assert Outer.__pydantic_serializer__.to_json(outer, serialize_as_any=False) == b'{"p":{"x":1}}'
    assert (
        Outer.__pydantic_serializer__.to_json(outer, serialize_as_any=True)
        == b'{"p":{"x":1,"y":"hopefully not a secret"}}'
    )


def test_serialize_with_recursive_models() -> None:
    class Node:
        next: Optional['Node'] = None
        value: int = 42

    schema = core_schema.definitions_schema(
        core_schema.definition_reference_schema('Node'),
        [
            core_schema.model_schema(
                Node,
                core_schema.model_fields_schema(
                    {
                        'value': core_schema.model_field(
                            core_schema.with_default_schema(core_schema.int_schema(), default=42)
                        ),
                        'next': core_schema.model_field(
                            core_schema.with_default_schema(
                                core_schema.nullable_schema(core_schema.definition_reference_schema('Node')),
                                default=None,
                            )
                        ),
                    }
                ),
                ref='Node',
            )
        ],
    )

    Node.__pydantic_core_schema__ = schema
    Node.__pydantic_validator__ = SchemaValidator(Node.__pydantic_core_schema__)
    Node.__pydantic_serializer__ = SchemaSerializer(Node.__pydantic_core_schema__)
    other = Node.__pydantic_validator__.validate_python({'next': {'value': 4}})

    assert Node.__pydantic_serializer__.to_python(other, serialize_as_any=False) == {
        'next': {'next': None, 'value': 4},
        'value': 42,
    }
    assert Node.__pydantic_serializer__.to_python(other, serialize_as_any=True) == {
        'next': {'next': None, 'value': 4},
        'value': 42,
    }


def test_serialize_with_custom_type_and_subclasses():
    class Node:
        x: int

    Node.__pydantic_core_schema__ = core_schema.model_schema(
        Node,
        core_schema.model_fields_schema(
            {
                'x': core_schema.model_field(core_schema.int_schema()),
            }
        ),
        ref='Node',
    )
    Node.__pydantic_validator__ = SchemaValidator(Node.__pydantic_core_schema__)
    Node.__pydantic_serializer__ = SchemaSerializer(Node.__pydantic_core_schema__)

    class NodeSubClass(Node):
        y: int

    NodeSubClass.__pydantic_core_schema__ = core_schema.model_schema(
        NodeSubClass,
        core_schema.model_fields_schema(
            {
                'x': core_schema.model_field(core_schema.int_schema()),
                'y': core_schema.model_field(core_schema.int_schema()),
            }
        ),
    )
    NodeSubClass.__pydantic_validator__ = SchemaValidator(NodeSubClass.__pydantic_core_schema__)
    NodeSubClass.__pydantic_serializer__ = SchemaSerializer(NodeSubClass.__pydantic_core_schema__)

    class CustomType:
        values: list[Node]

    CustomType.__pydantic_core_schema__ = core_schema.model_schema(
        CustomType,
        core_schema.definitions_schema(
            core_schema.model_fields_schema(
                {
                    'values': core_schema.model_field(
                        core_schema.list_schema(core_schema.definition_reference_schema('Node'))
                    ),
                }
            ),
            [
                Node.__pydantic_core_schema__,
            ],
        ),
    )
    CustomType.__pydantic_validator__ = SchemaValidator(CustomType.__pydantic_core_schema__)
    CustomType.__pydantic_serializer__ = SchemaSerializer(CustomType.__pydantic_core_schema__)

    value = CustomType.__pydantic_validator__.validate_python({'values': [{'x': 1}, {'x': 2}]})
    value.values.append(NodeSubClass.__pydantic_validator__.validate_python({'x': 3, 'y': 4}))

    assert CustomType.__pydantic_serializer__.to_python(value, serialize_as_any=False) == {
        'values': [{'x': 1}, {'x': 2}, {'x': 3}],
    }
    assert CustomType.__pydantic_serializer__.to_python(value, serialize_as_any=True) == {
        'values': [{'x': 1}, {'x': 2}, {'x': 3, 'y': 4}],
    }


def test_serialize_as_any_wrap_serializer_applied_once() -> None:
    # https://github.com/pydantic/pydantic/issues/11139

    class InnerModel:
        an_inner_field: int

    InnerModel.__pydantic_core_schema__ = core_schema.model_schema(
        InnerModel,
        core_schema.model_fields_schema({'an_inner_field': core_schema.model_field(core_schema.int_schema())}),
    )
    InnerModel.__pydantic_validator__ = SchemaValidator(InnerModel.__pydantic_core_schema__)
    InnerModel.__pydantic_serializer__ = SchemaSerializer(InnerModel.__pydantic_core_schema__)

    class MyModel:
        a_field: InnerModel

        def a_model_serializer(self, handler, info):
            return {k + '_wrapped': v for k, v in handler(self).items()}

    MyModel.__pydantic_core_schema__ = core_schema.model_schema(
        MyModel,
        core_schema.model_fields_schema({'a_field': core_schema.model_field(InnerModel.__pydantic_core_schema__)}),
        serialization=core_schema.wrap_serializer_function_ser_schema(
            MyModel.a_model_serializer,
            info_arg=True,
        ),
    )
    MyModel.__pydantic_validator__ = SchemaValidator(MyModel.__pydantic_core_schema__)
    MyModel.__pydantic_serializer__ = SchemaSerializer(MyModel.__pydantic_core_schema__)

    instance = MyModel.__pydantic_validator__.validate_python({'a_field': {'an_inner_field': 1}})
    assert MyModel.__pydantic_serializer__.to_python(instance, serialize_as_any=True) == {
        'a_field_wrapped': {'an_inner_field': 1},
    }



import re
import sys
from decimal import Decimal
from enum import Enum, IntEnum, IntFlag

import pytest

from pydantic_core import SchemaError, SchemaValidator, ValidationError, core_schema


def test_plain_enum():
    class MyEnum(Enum):
        a = 1
        b = 2

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values())))

    # debug(v)
    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python(MyEnum.b) is MyEnum.b
    assert v.validate_python(1) is MyEnum.a
    assert v.validate_python(2) is MyEnum.b

    assert v.validate_json('1') is MyEnum.a
    # assert v.validate_json('"1"') is MyEnum.a

    with pytest.raises(ValidationError, match=r'Input should be 1 or 2 \[type=enum, input_value=3, input_type=int\]'):
        v.validate_python(3)

    with pytest.raises(ValidationError, match=r"Input should be 1 or 2 \[type=enum, input_value='1', input_type=str\]"):
        v.validate_python('1')

    assert v.validate_python(MyEnum.a, strict=True) is MyEnum.a

    e = (
        'Input should be an instance of test_plain_enum.<locals>.MyEnum '
        '[type=is_instance_of, input_value=1, input_type=int]'
    )
    with pytest.raises(ValidationError, match=re.escape(e)):
        v.validate_python(1, strict=True)

    assert v.validate_json('1', strict=True) is MyEnum.a
    with pytest.raises(ValidationError, match='type=enum'):
        v.validate_json('"1"', strict=True)

    v_strict = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values()), strict=True))
    assert v_strict.validate_python(MyEnum.a) is MyEnum.a

    with pytest.raises(ValidationError, match=re.escape(e)):
        v_strict.validate_python(1, strict=True)

    v_strict_f = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values()), strict=True))
    assert v_strict_f.validate_python(MyEnum.a) is MyEnum.a

    with pytest.raises(ValidationError, match=re.escape(e)):
        v_strict_f.validate_python(1, strict=True)


def test_int_enum():
    class MyEnum(int, Enum):
        a = 1
        b = 2

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values()), sub_type='int'))

    # debug(v)
    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python(1) is MyEnum.a
    assert v.validate_python(1.0) is MyEnum.a
    assert v.validate_python('1') is MyEnum.a

    assert v.validate_json('1') is MyEnum.a
    assert v.validate_json('"1"') is MyEnum.a

    with pytest.raises(ValidationError, match=r'Input should be 1 or 2 \[type=enum, input_value=3, input_type=int\]'):
        v.validate_python(3)

    assert v.validate_python(MyEnum.a, strict=True) is MyEnum.a

    e = (
        'Input should be an instance of test_int_enum.<locals>.MyEnum '
        '[type=is_instance_of, input_value=1, input_type=int]'
    )
    with pytest.raises(ValidationError, match=re.escape(e)):
        v.validate_python(1, strict=True)

    assert v.validate_json('1', strict=True) is MyEnum.a

    with pytest.raises(ValidationError, match=r"Input should be 1 or 2 \[type=enum, input_value='1', input_type=str\]"):
        v.validate_json('"1"', strict=True)


def test_str_enum():
    class MyEnum(str, Enum):
        a = 'x'
        b = 'y'

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values()), sub_type='str'))

    # debug(v)
    assert v.validate_python('x') is MyEnum.a
    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python(b'x') is MyEnum.a

    assert v.validate_json('"x"') is MyEnum.a

    with pytest.raises(
        ValidationError, match=r"Input should be 'x' or 'y' \[type=enum, input_value='a', input_type=str\]"
    ):
        v.validate_python('a')

    assert v.validate_python(MyEnum.a, strict=True) is MyEnum.a

    e = (
        'Input should be an instance of test_str_enum.<locals>.MyEnum '
        "[type=is_instance_of, input_value='x', input_type=str]"
    )
    with pytest.raises(ValidationError, match=re.escape(e)):
        v.validate_python('x', strict=True)
    assert v.validate_json('"x"', strict=True) is MyEnum.a


def test_float_enum():
    class MyEnum(float, Enum):
        a = 1.5
        b = 2.5
        c = 3.0

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values()), sub_type='float'))

    # debug(v)
    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python(1.5) is MyEnum.a
    assert v.validate_python('1.5') is MyEnum.a
    assert v.validate_python(3) is MyEnum.c

    assert v.validate_json('1.5') is MyEnum.a
    assert v.validate_json('"1.5"') is MyEnum.a

    e = r'Input should be 1.5, 2.5 or 3.0 \[type=enum, input_value=4.0, input_type=float\]'
    with pytest.raises(ValidationError, match=e):
        v.validate_python(4.0)

    assert v.validate_python(MyEnum.a, strict=True) is MyEnum.a

    e = (
        'Input should be an instance of test_float_enum.<locals>.MyEnum '
        '[type=is_instance_of, input_value=1.5, input_type=float]'
    )
    with pytest.raises(ValidationError, match=re.escape(e)):
        v.validate_python(1.5, strict=True)

    assert v.validate_json('1.5', strict=True) is MyEnum.a

    with pytest.raises(ValidationError, match='type=enum'):
        v.validate_json('"3.0"', strict=True)


def test_enum_missing():
    class MyEnum(Enum):
        a = 1
        b = 2

        @classmethod
        def _missing_(cls, v):
            return cls.b

    assert MyEnum(1) is MyEnum.a
    assert MyEnum(2) is MyEnum.b
    assert MyEnum(3) is MyEnum.b

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values()), missing=MyEnum._missing_))

    # debug(v)
    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python(1) is MyEnum.a
    assert v.validate_python(2) is MyEnum.b
    assert v.validate_python(3) is MyEnum.b

    assert v.validate_json('1') is MyEnum.a
    assert v.validate_json('3') is MyEnum.b


def test_enum_missing_none():
    class MyEnum(Enum):
        a = 1
        b = 2

        @classmethod
        def _missing_(cls, v):
            return None

    assert MyEnum(1) is MyEnum.a
    assert MyEnum(2) is MyEnum.b
    with pytest.raises(ValueError, match='3 is not a valid'):
        MyEnum(3)

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values()), missing=MyEnum._missing_))

    # debug(v)
    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python(1) is MyEnum.a
    with pytest.raises(ValidationError, match=r'Input should be 1 or 2 \[type=enum, input_value=3, input_type=int\]'):
        v.validate_python(3)

    assert v.validate_json('1') is MyEnum.a
    with pytest.raises(ValidationError, match=r'Input should be 1 or 2 \[type=enum, input_value=3, input_type=int\]'):
        v.validate_json('3')


def test_enum_missing_wrong():
    class MyEnum(Enum):
        a = 1
        b = 2

        @classmethod
        def _missing_(cls, v):
            return 'foobar'

    assert MyEnum(1) is MyEnum.a
    assert MyEnum(2) is MyEnum.b
    # different error from pypy
    if sys.implementation.name == 'pypy':
        e = "returned 'foobar' instead of None or a valid member"
    else:
        e = "error in MyEnum._missing_: returned 'foobar' instead of None or a valid member"
    with pytest.raises(TypeError, match=e):
        MyEnum(3)

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values()), missing=MyEnum._missing_))
    with pytest.raises(TypeError, match=e):
        v.validate_python(3)


def test_enum_exactness():
    class MyEnum(int, Enum):
        a = 1
        b = 2

    v = SchemaValidator(
        core_schema.union_schema(
            [
                core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values()), missing=MyEnum._missing_),
                core_schema.int_schema(),
            ],
        )
    )
    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python(1) == 1
    assert v.validate_python(1) is not MyEnum.a


def test_plain_enum_lists():
    class MyEnum(Enum):
        a = [1]
        b = [2]

    assert MyEnum([1]) is MyEnum.a
    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values())))
    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python([1]) is MyEnum.a
    assert v.validate_python([2]) is MyEnum.b


def test_plain_enum_empty():
    class MyEnum(Enum):
        pass

    with pytest.raises(SchemaError, match='`members` should have length > 0'):
        SchemaValidator(core_schema.enum_schema(MyEnum, []))


def test_enum_with_str_subclass() -> None:
    class MyEnum(Enum):
        a = 'a'
        b = 'b'

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values())))

    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python('a') is MyEnum.a

    class MyStr(str):
        pass

    assert v.validate_python(MyStr('a')) is MyEnum.a
    with pytest.raises(ValidationError):
        v.validate_python(MyStr('a'), strict=True)


def test_enum_with_int_subclass() -> None:
    class MyEnum(Enum):
        a = 1
        b = 2

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values())))

    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python(1) is MyEnum.a

    class MyInt(int):
        pass

    assert v.validate_python(MyInt(1)) is MyEnum.a
    with pytest.raises(ValidationError):
        v.validate_python(MyInt(1), strict=True)


def test_validate_float_for_int_enum() -> None:
    class MyEnum(int, Enum):
        a = 1
        b = 2

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values())))

    assert v.validate_python(1.0) is MyEnum.a


def test_missing_error_converted_to_val_error() -> None:
    class MyFlags(IntFlag):
        OFF = 0
        ON = 1

    v = SchemaValidator(
        core_schema.with_default_schema(
            schema=core_schema.enum_schema(MyFlags, list(MyFlags.__members__.values())), default=MyFlags.OFF
        )
    )

    assert v.validate_python(MyFlags.OFF) is MyFlags.OFF
    assert v.validate_python(0) is MyFlags.OFF

    with pytest.raises(ValidationError):
        v.validate_python(None)


def test_big_int():
    class ColorEnum(IntEnum):
        GREEN = 1 << 63
        BLUE = 1 << 64

    v = SchemaValidator(
        core_schema.with_default_schema(schema=core_schema.enum_schema(ColorEnum, list(ColorEnum.__members__.values())))
    )

    assert v.validate_python(ColorEnum.GREEN) is ColorEnum.GREEN
    assert v.validate_python(1 << 63) is ColorEnum.GREEN


@pytest.mark.parametrize(
    'value',
    [-1, 0, 1],
)
def test_enum_int_validation_should_succeed_for_decimal(value: int):
    class MyEnum(Enum):
        VALUE = value

    class MyIntEnum(IntEnum):
        VALUE = value

    v = SchemaValidator(
        core_schema.with_default_schema(
            schema=core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values())),
            default=MyEnum.VALUE,
        )
    )

    v_int = SchemaValidator(
        core_schema.with_default_schema(
            schema=core_schema.enum_schema(MyIntEnum, list(MyIntEnum.__members__.values())),
            default=MyIntEnum.VALUE,
        )
    )

    assert v.validate_python(Decimal(value)) is MyEnum.VALUE
    assert v.validate_python(Decimal(float(value))) is MyEnum.VALUE
    assert v_int.validate_python(Decimal(value)) is MyIntEnum.VALUE
    assert v_int.validate_python(Decimal(float(value))) is MyIntEnum.VALUE


@pytest.mark.skipif(
    sys.version_info >= (3, 13),
    reason='Python 3.13+ enum initialization is different, see https://github.com/python/cpython/blob/ec610069637d56101896803a70d418a89afe0b4b/Lib/enum.py#L1159-L1163',
)
def test_enum_int_validation_should_succeed_for_custom_type():
    class AnyWrapper:
        def __init__(self, value):
            self.value = value

        def __eq__(self, other: object) -> bool:
            return self.value == other

    class MyEnum(Enum):
        VALUE = 999
        SECOND_VALUE = 1000000
        THIRD_VALUE = 'Py03'

    v = SchemaValidator(
        core_schema.with_default_schema(
            schema=core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values())),
            default=MyEnum.VALUE,
        )
    )

    assert v.validate_python(AnyWrapper(999)) is MyEnum.VALUE
    assert v.validate_python(AnyWrapper(1000000)) is MyEnum.SECOND_VALUE
    assert v.validate_python(AnyWrapper('Py03')) is MyEnum.THIRD_VALUE


def test_enum_str_validation_should_fail_for_decimal_when_expecting_str_value():
    class MyEnum(Enum):
        VALUE = '1'

    v = SchemaValidator(
        core_schema.with_default_schema(
            schema=core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values())),
            default=MyEnum.VALUE,
        )
    )

    with pytest.raises(ValidationError):
        v.validate_python(Decimal(1))


def test_enum_int_validation_should_fail_for_incorrect_decimal_value():
    class MyEnum(Enum):
        VALUE = 1

    v = SchemaValidator(
        core_schema.with_default_schema(
            schema=core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values())),
            default=MyEnum.VALUE,
        )
    )

    with pytest.raises(ValidationError):
        v.validate_python(Decimal(2))

    with pytest.raises(ValidationError):
        v.validate_python((1, 2))

    with pytest.raises(ValidationError):
        v.validate_python(Decimal(1.1))


def test_enum_int_validation_should_fail_for_plain_type_without_eq_checking():
    class MyEnum(Enum):
        VALUE = 1

    class MyClass:
        def __init__(self, value):
            self.value = value

    v = SchemaValidator(
        core_schema.with_default_schema(
            schema=core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values())),
            default=MyEnum.VALUE,
        )
    )

    with pytest.raises(ValidationError):
        v.validate_python(MyClass(1))


def support_custom_new_method() -> None:
    """Demonstrates support for custom new methods, as well as conceptually, multi-value enums without dependency on a 3rd party lib for testing."""

    class Animal(Enum):
        CAT = 'cat', 'meow'
        DOG = 'dog', 'woof'

        def __new__(cls, species: str, sound: str):
            obj = object.__new__(cls)

            obj._value_ = species
            obj._all_values = (species, sound)

            obj.species = species
            obj.sound = sound

            cls._value2member_map_[sound] = obj

            return obj

    v = SchemaValidator(core_schema.enum_schema(Animal, list(Animal.__members__.values())))
    assert v.validate_python('cat') is Animal.CAT
    assert v.validate_python('meow') is Animal.CAT
    assert v.validate_python('dog') is Animal.DOG
    assert v.validate_python('woof') is Animal.DOG


def test_strict_enum_wrap_json() -> None:
    """https://github.com/pydantic/pydantic/issues/11070"""

    class Animal(str, Enum):
        CAT = 'cat'
        DOG = 'dog'

    schema = core_schema.no_info_wrap_validator_function(
        lambda v, handler: handler(v),
        core_schema.enum_schema(Animal, list(Animal.__members__.values()), sub_type='str', strict=True),
    )
    v = SchemaValidator(schema)

    assert v.validate_python(Animal.CAT) == Animal.CAT
    assert v.validate_json('"dog"') == Animal.DOG



from typing import Union

from pydantic_core import SchemaSerializer, SchemaValidator, core_schema


def test_prebuilt_val_and_ser_used() -> None:
    class InnerModel:
        x: int

    inner_schema = core_schema.model_schema(
        InnerModel,
        schema=core_schema.model_fields_schema(
            {'x': core_schema.model_field(schema=core_schema.int_schema())},
        ),
    )

    inner_validator = SchemaValidator(inner_schema)
    inner_serializer = SchemaSerializer(inner_schema)
    InnerModel.__pydantic_complete__ = True  # pyright: ignore[reportAttributeAccessIssue]
    InnerModel.__pydantic_validator__ = inner_validator  # pyright: ignore[reportAttributeAccessIssue]
    InnerModel.__pydantic_serializer__ = inner_serializer  # pyright: ignore[reportAttributeAccessIssue]

    class OuterModel:
        inner: InnerModel

    outer_schema = core_schema.model_schema(
        OuterModel,
        schema=core_schema.model_fields_schema(
            {
                'inner': core_schema.model_field(
                    schema=core_schema.model_schema(
                        InnerModel,
                        schema=core_schema.model_fields_schema(
                            # note, we use str schema here even though that's incorrect
                            # in order to verify that the prebuilt validator is used
                            # off of InnerModel with the correct int schema, not this str schema
                            {'x': core_schema.model_field(schema=core_schema.str_schema())},
                        ),
                    )
                )
            }
        ),
    )

    outer_validator = SchemaValidator(outer_schema)
    outer_serializer = SchemaSerializer(outer_schema)

    result = outer_validator.validate_python({'inner': {'x': 1}})
    assert result.inner.x == 1
    assert outer_serializer.to_python(result) == {'inner': {'x': 1}}


def test_prebuilt_not_used_for_wrap_serializer_functions() -> None:
    class InnerModel:
        x: str

        def __init__(self, x: str) -> None:
            self.x = x

    def serialize_inner(v: InnerModel, serializer) -> Union[dict[str, str], str]:
        v.x = v.x + ' modified'
        return serializer(v)

    inner_schema = core_schema.model_schema(
        InnerModel,
        schema=core_schema.model_fields_schema(
            {'x': core_schema.model_field(schema=core_schema.str_schema())},
        ),
        serialization=core_schema.wrap_serializer_function_ser_schema(serialize_inner),
    )

    inner_serializer = SchemaSerializer(inner_schema)
    InnerModel.__pydantic_complete__ = True  # pyright: ignore[reportAttributeAccessIssue]
    InnerModel.__pydantic_serializer__ = inner_serializer  # pyright: ignore[reportAttributeAccessIssue]

    class OuterModel:
        inner: InnerModel

        def __init__(self, inner: InnerModel) -> None:
            self.inner = inner

    outer_schema = core_schema.model_schema(
        OuterModel,
        schema=core_schema.model_fields_schema(
            {
                'inner': core_schema.model_field(
                    schema=core_schema.model_schema(
                        InnerModel,
                        schema=core_schema.model_fields_schema(
                            # note, we use a simple str schema (with no custom serialization)
                            # in order to verify that the prebuilt serializer from InnerModel is not used
                            {'x': core_schema.model_field(schema=core_schema.str_schema())},
                        ),
                    )
                )
            }
        ),
    )

    outer_serializer = SchemaSerializer(outer_schema)

    # the custom serialization function does apply for the inner model
    inner_instance = InnerModel(x='hello')
    assert inner_serializer.to_python(inner_instance) == {'x': 'hello modified'}

    # but the outer model doesn't reuse the custom wrap serializer function, so we see simple str ser
    outer_instance = OuterModel(inner=InnerModel(x='hello'))
    assert outer_serializer.to_python(outer_instance) == {'inner': {'x': 'hello'}}


def test_prebuilt_not_used_for_wrap_validator_functions() -> None:
    class InnerModel:
        x: str

        def __init__(self, x: str) -> None:
            self.x = x

    def validate_inner(data, validator) -> InnerModel:
        data['x'] = data['x'] + ' modified'
        return validator(data)

    inner_schema = core_schema.no_info_wrap_validator_function(
        validate_inner,
        core_schema.model_schema(
            InnerModel,
            schema=core_schema.model_fields_schema(
                {'x': core_schema.model_field(schema=core_schema.str_schema())},
            ),
        ),
    )

    inner_validator = SchemaValidator(inner_schema)
    InnerModel.__pydantic_complete__ = True  # pyright: ignore[reportAttributeAccessIssue]
    InnerModel.__pydantic_validator__ = inner_validator  # pyright: ignore[reportAttributeAccessIssue]

    class OuterModel:
        inner: InnerModel

        def __init__(self, inner: InnerModel) -> None:
            self.inner = inner

    outer_schema = core_schema.model_schema(
        OuterModel,
        schema=core_schema.model_fields_schema(
            {
                'inner': core_schema.model_field(
                    schema=core_schema.model_schema(
                        InnerModel,
                        schema=core_schema.model_fields_schema(
                            # note, we use a simple str schema (with no custom validation)
                            # in order to verify that the prebuilt validator from InnerModel is not used
                            {'x': core_schema.model_field(schema=core_schema.str_schema())},
                        ),
                    )
                )
            }
        ),
    )

    outer_validator = SchemaValidator(outer_schema)

    # the custom validation function does apply for the inner model
    result_inner = inner_validator.validate_python({'x': 'hello'})
    assert result_inner.x == 'hello modified'

    # but the outer model doesn't reuse the custom wrap validator function, so we see simple str val
    result_outer = outer_validator.validate_python({'inner': {'x': 'hello'}})
    assert result_outer.inner.x == 'hello'


def test_prebuilt_not_used_for_after_validator_functions() -> None:
    class InnerModel:
        x: str

        def __init__(self, x: str) -> None:
            self.x = x

    def validate_after(self) -> InnerModel:
        self.x = self.x + ' modified'
        return self

    inner_schema = core_schema.no_info_after_validator_function(
        validate_after,
        core_schema.model_schema(
            InnerModel,
            schema=core_schema.model_fields_schema(
                {'x': core_schema.model_field(schema=core_schema.str_schema())},
            ),
        ),
    )

    inner_validator = SchemaValidator(inner_schema)
    InnerModel.__pydantic_complete__ = True  # pyright: ignore[reportAttributeAccessIssue]
    InnerModel.__pydantic_validator__ = inner_validator  # pyright: ignore[reportAttributeAccessIssue]

    class OuterModel:
        inner: InnerModel

        def __init__(self, inner: InnerModel) -> None:
            self.inner = inner

    outer_schema = core_schema.model_schema(
        OuterModel,
        schema=core_schema.model_fields_schema(
            {
                'inner': core_schema.model_field(
                    schema=core_schema.model_schema(
                        InnerModel,
                        schema=core_schema.model_fields_schema(
                            # note, we use a simple str schema (with no custom validation)
                            # in order to verify that the prebuilt validator from InnerModel is not used
                            {'x': core_schema.model_field(schema=core_schema.str_schema())},
                        ),
                    )
                )
            }
        ),
    )

    outer_validator = SchemaValidator(outer_schema)

    # the custom validation function does apply for the inner model
    result_inner = inner_validator.validate_python({'x': 'hello'})
    assert result_inner.x == 'hello modified'

    # but the outer model doesn't reuse the custom after validator function, so we see simple str val
    result_outer = outer_validator.validate_python({'inner': {'x': 'hello'}})
    assert result_outer.inner.x == 'hello'


def test_reuse_plain_serializer_ok() -> None:
    class InnerModel:
        x: str

        def __init__(self, x: str) -> None:
            self.x = x

    def serialize_inner(v: InnerModel) -> str:
        return v.x + ' modified'

    inner_schema = core_schema.model_schema(
        InnerModel,
        schema=core_schema.model_fields_schema(
            {'x': core_schema.model_field(schema=core_schema.str_schema())},
        ),
        serialization=core_schema.plain_serializer_function_ser_schema(serialize_inner),
    )

    inner_serializer = SchemaSerializer(inner_schema)
    InnerModel.__pydantic_complete__ = True  # pyright: ignore[reportAttributeAccessIssue]
    InnerModel.__pydantic_serializer__ = inner_serializer  # pyright: ignore[reportAttributeAccessIssue]

    class OuterModel:
        inner: InnerModel

        def __init__(self, inner: InnerModel) -> None:
            self.inner = inner

    outer_schema = core_schema.model_schema(
        OuterModel,
        schema=core_schema.model_fields_schema(
            {
                'inner': core_schema.model_field(
                    schema=core_schema.model_schema(
                        InnerModel,
                        schema=core_schema.model_fields_schema(
                            # note, we use a simple str schema (with no custom serialization)
                            # in order to verify that the prebuilt serializer from InnerModel is used instead
                            {'x': core_schema.model_field(schema=core_schema.str_schema())},
                        ),
                    )
                )
            }
        ),
    )

    outer_serializer = SchemaSerializer(outer_schema)

    # the custom serialization function does apply for the inner model
    inner_instance = InnerModel(x='hello')
    assert inner_serializer.to_python(inner_instance) == 'hello modified'
    assert 'FunctionPlainSerializer' in repr(inner_serializer)

    # the custom ser function applies for the outer model as well, a plain serializer is permitted as a prebuilt candidate
    outer_instance = OuterModel(inner=InnerModel(x='hello'))
    assert outer_serializer.to_python(outer_instance) == {'inner': 'hello modified'}
    assert 'PrebuiltSerializer' in repr(outer_serializer)


def test_reuse_plain_validator_ok() -> None:
    class InnerModel:
        x: str

        def __init__(self, x: str) -> None:
            self.x = x

    def validate_inner(data) -> InnerModel:
        data['x'] = data['x'] + ' modified'
        return InnerModel(**data)

    inner_schema = core_schema.no_info_plain_validator_function(validate_inner)

    inner_validator = SchemaValidator(inner_schema)
    InnerModel.__pydantic_complete__ = True  # pyright: ignore[reportAttributeAccessIssue]
    InnerModel.__pydantic_validator__ = inner_validator  # pyright: ignore[reportAttributeAccessIssue]

    class OuterModel:
        inner: InnerModel

        def __init__(self, inner: InnerModel) -> None:
            self.inner = inner

    outer_schema = core_schema.model_schema(
        OuterModel,
        schema=core_schema.model_fields_schema(
            {
                'inner': core_schema.model_field(
                    schema=core_schema.model_schema(
                        InnerModel,
                        schema=core_schema.model_fields_schema(
                            # note, we use a simple str schema (with no custom validation)
                            # in order to verify that the prebuilt validator from InnerModel is used instead
                            {'x': core_schema.model_field(schema=core_schema.str_schema())},
                        ),
                    )
                )
            }
        ),
    )

    outer_validator = SchemaValidator(outer_schema)

    # the custom validation function does apply for the inner model
    result_inner = inner_validator.validate_python({'x': 'hello'})
    assert result_inner.x == 'hello modified'
    assert 'FunctionPlainValidator' in repr(inner_validator)

    # the custom validation function does apply for the outer model as well, a plain validator is permitted as a prebuilt candidate
    result_outer = outer_validator.validate_python({'inner': {'x': 'hello'}})
    assert result_outer.inner.x == 'hello modified'
    assert 'PrebuiltValidator' in repr(outer_validator)


def test_reuse_before_validator_ok() -> None:
    class InnerModel:
        x: str

        def __init__(self, x: str) -> None:
            self.x = x

    def validate_before(data) -> dict:
        data['x'] = data['x'] + ' modified'
        return data

    inner_schema = core_schema.no_info_before_validator_function(
        validate_before,
        core_schema.model_schema(
            InnerModel,
            schema=core_schema.model_fields_schema(
                {'x': core_schema.model_field(schema=core_schema.str_schema())},
            ),
        ),
    )

    inner_validator = SchemaValidator(inner_schema)
    InnerModel.__pydantic_complete__ = True  # pyright: ignore[reportAttributeAccessIssue]
    InnerModel.__pydantic_validator__ = inner_validator  # pyright: ignore[reportAttributeAccessIssue]

    class OuterModel:
        inner: InnerModel

        def __init__(self, inner: InnerModel) -> None:
            self.inner = inner

    outer_schema = core_schema.model_schema(
        OuterModel,
        schema=core_schema.model_fields_schema(
            {
                'inner': core_schema.model_field(
                    schema=core_schema.model_schema(
                        InnerModel,
                        schema=core_schema.model_fields_schema(
                            # note, we use a simple str schema (with no custom validation)
                            # in order to verify that the prebuilt validator from InnerModel is used instead
                            {'x': core_schema.model_field(schema=core_schema.str_schema())},
                        ),
                    )
                )
            }
        ),
    )

    outer_validator = SchemaValidator(outer_schema)
    print(inner_validator)
    print(outer_validator)

    # the custom validation function does apply for the inner model
    result_inner = inner_validator.validate_python({'x': 'hello'})
    assert result_inner.x == 'hello modified'
    assert 'FunctionBeforeValidator' in repr(inner_validator)

    # the custom validation function does apply for the outer model as well, a before validator is permitted as a prebuilt candidate
    result_outer = outer_validator.validate_python({'inner': {'x': 'hello'}})
    assert result_outer.inner.x == 'hello modified'
    assert 'PrebuiltValidator' in repr(outer_validator)



import json
import platform
import re

import pytest
from dirty_equals import IsFloatNan, IsList

import pydantic_core
from pydantic_core import (
    CoreConfig,
    PydanticSerializationError,
    SchemaSerializer,
    SchemaValidator,
    ValidationError,
    core_schema,
    from_json,
    to_json,
    to_jsonable_python,
)

from .conftest import Err


@pytest.mark.parametrize(
    'input_value,output_value',
    [('false', False), ('true', True), ('0', False), ('1', True), ('"yes"', True), ('"no"', False)],
)
def test_bool(input_value, output_value):
    v = SchemaValidator(core_schema.bool_schema())
    assert v.validate_json(input_value) == output_value


@pytest.mark.parametrize('input_value', ['[1, 2, 3]', b'[1, 2, 3]', bytearray(b'[1, 2, 3]')])
def test_input_types(input_value):
    v = SchemaValidator(core_schema.list_schema(items_schema=core_schema.int_schema()))
    assert v.validate_json(input_value) == [1, 2, 3]


def test_input_type_invalid():
    v = SchemaValidator(core_schema.list_schema(items_schema=core_schema.int_schema()))
    with pytest.raises(ValidationError, match=r'JSON input should be string, bytes or bytearray \[type=json_type,'):
        v.validate_json([])


def test_null():
    assert SchemaValidator(core_schema.none_schema()).validate_json('null') is None


def test_str():
    s = SchemaValidator(core_schema.str_schema())
    assert s.validate_json('"foobar"') == 'foobar'
    with pytest.raises(ValidationError, match=r'Input should be a valid string \[type=string_type,'):
        s.validate_json('false')
    with pytest.raises(ValidationError, match=r'Input should be a valid string \[type=string_type,'):
        s.validate_json('123')


def test_bytes():
    s = SchemaValidator(core_schema.bytes_schema())
    assert s.validate_json('"foobar"') == b'foobar'
    with pytest.raises(ValidationError, match=r'Input should be a valid bytes \[type=bytes_type,'):
        s.validate_json('false')
    with pytest.raises(ValidationError, match=r'Input should be a valid bytes \[type=bytes_type,'):
        s.validate_json('123')


# A number well outside of i64 range
_BIG_NUMBER_STR = '1' + ('0' * 40)


@pytest.mark.parametrize(
    'input_value,expected',
    [
        ('123', 123),
        ('"123"', 123),
        ('123.0', 123),
        ('"123.0"', 123),
        (_BIG_NUMBER_STR, int(_BIG_NUMBER_STR)),
        ('123.4', Err('Input should be a valid integer, got a number with a fractional part [type=int_from_float,')),
        ('"123.4"', Err('Input should be a valid integer, unable to parse string as an integer [type=int_parsing,')),
        ('"string"', Err('Input should be a valid integer, unable to parse string as an integer [type=int_parsing,')),
    ],
)
def test_int(input_value, expected):
    v = SchemaValidator(core_schema.int_schema())
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_json(input_value)
    else:
        assert v.validate_json(input_value) == expected


@pytest.mark.parametrize(
    'input_value,expected',
    [
        ('123.4', 123.4),
        ('123.0', 123.0),
        ('123', 123.0),
        ('"123.4"', 123.4),
        ('"123.0"', 123.0),
        ('"123"', 123.0),
        ('"string"', Err('Input should be a valid number, unable to parse string as a number [type=float_parsing,')),
    ],
)
def test_float(input_value, expected):
    v = SchemaValidator(core_schema.float_schema())
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_json(input_value)
    else:
        assert v.validate_json(input_value) == expected


def test_typed_dict():
    v = SchemaValidator(
        core_schema.typed_dict_schema(
            fields={
                'field_a': core_schema.typed_dict_field(schema=core_schema.str_schema()),
                'field_b': core_schema.typed_dict_field(schema=core_schema.int_schema()),
            }
        )
    )

    # language=json
    input_str = '{"field_a": "abc", "field_b": 1}'
    assert v.validate_json(input_str) == {'field_a': 'abc', 'field_b': 1}
    # language=json
    input_str = '{"field_a": "a", "field_a": "b", "field_b": 1}'
    assert v.validate_json(input_str) == {'field_a': 'b', 'field_b': 1}
    assert v.validate_json(input_str) == {'field_a': 'b', 'field_b': 1}


def test_float_no_remainder():
    v = SchemaValidator(core_schema.int_schema())
    assert v.validate_json('123.0') == 123


def test_error_loc():
    v = SchemaValidator(
        core_schema.typed_dict_schema(
            fields={
                'field_a': core_schema.typed_dict_field(
                    schema=core_schema.list_schema(items_schema=core_schema.int_schema())
                )
            },
            extras_schema=core_schema.int_schema(),
            extra_behavior='allow',
        )
    )

    # assert v.validate_json('{"field_a": [1, 2, "3"]}') == ({'field_a': [1, 2, 3]}, {'field_a'})

    with pytest.raises(ValidationError) as exc_info:
        v.validate_json('{"field_a": [1, 2, "wrong"]}')
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'int_parsing',
            'loc': ('field_a', 2),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'wrong',
        }
    ]


def test_dict():
    v = SchemaValidator(
        core_schema.dict_schema(keys_schema=core_schema.int_schema(), values_schema=core_schema.int_schema())
    )
    assert v.validate_json('{"1": 2, "3": 4}') == {1: 2, 3: 4}

    # duplicate keys, the last value wins, like with python
    assert json.loads('{"1": 1, "1": 2}') == {'1': 2}
    assert v.validate_json('{"1": 1, "1": 2}') == {1: 2}


def test_dict_any_value():
    v = SchemaValidator(core_schema.dict_schema(keys_schema=core_schema.str_schema()))
    assert v.validate_json('{"1": 1, "2": "a", "3": null}') == {'1': 1, '2': 'a', '3': None}


def test_json_invalid():
    v = SchemaValidator(core_schema.bool_schema())

    with pytest.raises(ValidationError) as exc_info:
        v.validate_json('"foobar')
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'json_invalid',
            'loc': (),
            'msg': 'Invalid JSON: EOF while parsing a string at line 1 column 7',
            'input': '"foobar',
            'ctx': {'error': 'EOF while parsing a string at line 1 column 7'},
        }
    ]
    with pytest.raises(ValidationError) as exc_info:
        v.validate_json('[1,\n2,\n3,]')
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'json_invalid',
            'loc': (),
            'msg': 'Invalid JSON: trailing comma at line 3 column 3',
            'input': '[1,\n2,\n3,]',
            'ctx': {'error': 'trailing comma at line 3 column 3'},
        }
    ]


class Foobar:
    def __str__(self):
        return 'Foobar.__str__'


def fallback_func(v):
    return f'fallback:{type(v).__name__}'


def test_to_json():
    assert to_json([1, 2]) == b'[1,2]'
    assert to_json([1, 2], indent=2) == b'[\n  1,\n  2\n]'
    assert to_json([1, b'x']) == b'[1,"x"]'
    assert to_json(['à', 'é']).decode('utf-8') == '["à","é"]'
    assert to_json(['à', 'é'], indent=2).decode('utf-8') == '[\n  "à",\n  "é"\n]'
    assert to_json(['à', 'é'], indent=2, ensure_ascii=True).decode('utf-8') == '[\n  "\\u00e0",\n  "\\u00e9"\n]'

    # kwargs required
    with pytest.raises(TypeError, match=r'to_json\(\) takes 1 positional arguments but 2 were given'):
        to_json([1, 2], 2)


def test_to_json_fallback():
    with pytest.raises(PydanticSerializationError, match=r'Unable to serialize unknown type: <.+\.Foobar'):
        to_json(Foobar())

    assert to_json(Foobar(), serialize_unknown=True) == b'"Foobar.__str__"'
    assert to_json(Foobar(), serialize_unknown=True, fallback=fallback_func) == b'"fallback:Foobar"'
    assert to_json(Foobar(), fallback=fallback_func) == b'"fallback:Foobar"'


def test_to_jsonable_python():
    assert to_jsonable_python([1, 2]) == [1, 2]
    assert to_jsonable_python({1, 2}) == IsList(1, 2, check_order=False)
    assert to_jsonable_python([1, b'x']) == [1, 'x']
    assert to_jsonable_python([0, 1, 2, 3, 4], exclude={1, 3}) == [0, 2, 4]


def test_to_jsonable_python_fallback():
    with pytest.raises(PydanticSerializationError, match=r'Unable to serialize unknown type: <.+\.Foobar'):
        to_jsonable_python(Foobar())

    assert to_jsonable_python(Foobar(), serialize_unknown=True) == 'Foobar.__str__'
    assert to_jsonable_python(Foobar(), serialize_unknown=True, fallback=fallback_func) == 'fallback:Foobar'
    assert to_jsonable_python(Foobar(), fallback=fallback_func) == 'fallback:Foobar'


def test_to_jsonable_python_schema_serializer():
    class Foobar:
        def __init__(self, my_foo: int, my_inners: list['Foobar']):
            self.my_foo = my_foo
            self.my_inners = my_inners

    # force a recursive model to ensure we exercise the transfer of definitions from the loaded
    # serializer
    c = core_schema.definitions_schema(
        core_schema.definition_reference_schema(schema_ref='foobar'),
        [
            core_schema.model_schema(
                Foobar,
                core_schema.typed_dict_schema(
                    {
                        'my_foo': core_schema.typed_dict_field(core_schema.int_schema(), serialization_alias='myFoo'),
                        'my_inners': core_schema.typed_dict_field(
                            core_schema.list_schema(core_schema.definition_reference_schema('foobar')),
                            serialization_alias='myInners',
                        ),
                    }
                ),
                ref='foobar',
            )
        ],
    )
    v = SchemaValidator(c)
    s = SchemaSerializer(c)

    Foobar.__pydantic_validator__ = v
    Foobar.__pydantic_serializer__ = s

    instance = Foobar(my_foo=1, my_inners=[Foobar(my_foo=2, my_inners=[])])
    assert to_jsonable_python(instance, by_alias=True) == {'myFoo': 1, 'myInners': [{'myFoo': 2, 'myInners': []}]}
    assert to_jsonable_python(instance, by_alias=False) == {'my_foo': 1, 'my_inners': [{'my_foo': 2, 'my_inners': []}]}
    assert to_json(instance, by_alias=True) == b'{"myFoo":1,"myInners":[{"myFoo":2,"myInners":[]}]}'
    assert to_json(instance, by_alias=False) == b'{"my_foo":1,"my_inners":[{"my_foo":2,"my_inners":[]}]}'


def test_cycle_same():
    def fallback_func_passthrough(obj):
        return obj

    f = Foobar()

    with pytest.raises(ValueError, match=r'Circular reference detected \(id repeated\)'):
        to_jsonable_python(f, fallback=fallback_func_passthrough)

    with pytest.raises(ValueError, match=r'Circular reference detected \(id repeated\)'):
        to_json(f, fallback=fallback_func_passthrough)


@pytest.mark.skipif(
    platform.python_implementation() == 'PyPy' and pydantic_core._pydantic_core.build_profile == 'debug',
    reason='PyPy does not have enough stack space for Rust debug builds to recurse very deep',
)
def test_cycle_change():
    def fallback_func_change_id(obj):
        return Foobar()

    f = Foobar()

    with pytest.raises(ValueError, match=r'Circular reference detected \(depth exceeded\)'):
        to_jsonable_python(f, fallback=fallback_func_change_id)

    with pytest.raises(ValueError, match=r'Circular reference detected \(depth exceeded\)'):
        to_json(f, fallback=fallback_func_change_id)


class FoobarHash:
    def __str__(self):
        return 'Foobar.__str__'

    def __hash__(self):
        return 1


def test_json_key_fallback():
    x = {FoobarHash(): 1}

    assert to_jsonable_python(x, serialize_unknown=True) == {'Foobar.__str__': 1}
    assert to_jsonable_python(x, fallback=fallback_func) == {'fallback:FoobarHash': 1}
    assert to_json(x, serialize_unknown=True) == b'{"Foobar.__str__":1}'
    assert to_json(x, fallback=fallback_func) == b'{"fallback:FoobarHash":1}'


class BedReprMeta(type):
    def __repr__(self):
        raise ValueError('bad repr')


class BadRepr(metaclass=BedReprMeta):
    def __repr__(self):
        raise ValueError('bad repr')

    def __hash__(self):
        return 1


def test_bad_repr():
    b = BadRepr()

    error_msg = '^Unable to serialize unknown type: <unprintable BedReprMeta object>$'
    with pytest.raises(PydanticSerializationError, match=error_msg):
        to_jsonable_python(b)

    assert to_jsonable_python(b, serialize_unknown=True) == '<Unserializable BadRepr object>'

    with pytest.raises(PydanticSerializationError, match=error_msg):
        to_json(b)

    assert to_json(b, serialize_unknown=True) == b'"<Unserializable BadRepr object>"'


def test_inf_nan_allow():
    v = SchemaValidator(core_schema.float_schema(allow_inf_nan=True))
    assert v.validate_json('Infinity') == float('inf')
    assert v.validate_json('-Infinity') == float('-inf')
    assert v.validate_json('NaN') == IsFloatNan()


def test_partial_parse():
    with pytest.raises(ValueError, match='EOF while parsing a string at line 1 column 15'):
        from_json('["aa", "bb", "c')
    assert from_json('["aa", "bb", "c', allow_partial=True) == ['aa', 'bb']

    with pytest.raises(ValueError, match='EOF while parsing a string at line 1 column 15'):
        from_json(b'["aa", "bb", "c')
    assert from_json(b'["aa", "bb", "c', allow_partial=True) == ['aa', 'bb']


def test_json_bytes_base64_round_trip():
    data = b'\xd8\x07\xc1Tx$\x91F%\xf3\xf3I\xca\xd8@\x0c\xee\xc3\xab\xff\x7f\xd3\xcd\xcd\xf9\xc2\x10\xe4\xa1\xb01e'
    encoded_std = b'"2AfBVHgkkUYl8/NJythADO7Dq/9/083N+cIQ5KGwMWU="'
    encoded_url = b'"2AfBVHgkkUYl8_NJythADO7Dq_9_083N-cIQ5KGwMWU="'
    assert to_json(data, bytes_mode='base64') == encoded_url

    v = SchemaValidator(core_schema.bytes_schema(), config=CoreConfig(val_json_bytes='base64'))
    assert v.validate_json(encoded_url) == data
    assert v.validate_json(encoded_std) == data

    with pytest.raises(ValidationError) as exc:
        v.validate_json('"wrong!"')
    [details] = exc.value.errors()
    assert details['type'] == 'bytes_invalid_encoding'

    assert to_json({'key': data}, bytes_mode='base64') == b'{"key":' + encoded_url + b'}'
    v = SchemaValidator(
        core_schema.dict_schema(keys_schema=core_schema.str_schema(), values_schema=core_schema.bytes_schema()),
        config=CoreConfig(val_json_bytes='base64'),
    )
    assert v.validate_json(b'{"key":' + encoded_url + b'}') == {'key': data}


def test_json_bytes_base64_no_padding():
    v = SchemaValidator(core_schema.bytes_schema(), config=CoreConfig(val_json_bytes='base64'))
    base_64_without_padding = 'bm8tcGFkZGluZw'
    assert v.validate_json(json.dumps(base_64_without_padding)) == b'no-padding'


def test_json_bytes_base64_invalid():
    v = SchemaValidator(core_schema.bytes_schema(), config=CoreConfig(val_json_bytes='base64'))
    wrong_input = 'wrong!'
    with pytest.raises(ValidationError) as exc_info:
        v.validate_json(json.dumps(wrong_input))
    assert exc_info.value.errors(include_url=False, include_context=False) == [
        {
            'type': 'bytes_invalid_encoding',
            'loc': (),
            'msg': f'Data should be valid base64: Invalid symbol {ord("!")}, offset {len(wrong_input) - 1}.',
            'input': wrong_input,
        }
    ]


def test_json_bytes_hex_round_trip():
    data = b'hello'
    encoded = b'"68656c6c6f"'
    assert to_json(data, bytes_mode='hex') == encoded

    v = SchemaValidator(core_schema.bytes_schema(), config=CoreConfig(val_json_bytes='hex'))
    assert v.validate_json(encoded) == data

    assert to_json({'key': data}, bytes_mode='hex') == b'{"key":"68656c6c6f"}'
    v = SchemaValidator(
        core_schema.dict_schema(keys_schema=core_schema.str_schema(), values_schema=core_schema.bytes_schema()),
        config=CoreConfig(val_json_bytes='hex'),
    )
    assert v.validate_json('{"key":"68656c6c6f"}') == {'key': data}


def test_json_bytes_hex_invalid():
    v = SchemaValidator(core_schema.bytes_schema(), config=CoreConfig(val_json_bytes='hex'))

    wrong_input = 'a'
    with pytest.raises(ValidationError) as exc_info:
        v.validate_json(json.dumps(wrong_input))
    assert exc_info.value.errors(include_url=False, include_context=False) == [
        {
            'type': 'bytes_invalid_encoding',
            'loc': (),
            'msg': 'Data should be valid hex: Odd number of digits',
            'input': wrong_input,
        }
    ]

    wrong_input = 'ag'
    with pytest.raises(ValidationError) as exc_info:
        v.validate_json(json.dumps(wrong_input))
    assert exc_info.value.errors(include_url=False, include_context=False) == [
        {
            'type': 'bytes_invalid_encoding',
            'loc': (),
            'msg': "Data should be valid hex: Invalid character 'g' at position 1",
            'input': wrong_input,
        }
    ]
